'''Test satellite trail detection and masking using findsat_mrt.'''
import os

import pytest
from astropy.io.fits import FITSDiff
from astropy.utils.data import get_pkg_data_filename

from acstools.findsat_mrt import WfcWrapper

pytest.importorskip("skimage")


def test_WfcWrapper(tmp_path):
    """Identify and mask trails in WFC extension 4."""
    timeout = 30
    rootname = 'jc8m32j5q'
    inputfile = get_pkg_data_filename(
        os.path.join('data', 'input', f'{rootname}_flc.fits'),
        package='acstools.tests', show_progress=False, remote_timeout=timeout)
    desired = get_pkg_data_filename(
        os.path.join('data', 'truth', f'{rootname}_flc_mask_ref.fits'),
        package='acstools.tests', show_progress=False, remote_timeout=timeout)
    actual = tmp_path / f'{rootname}_flc_mask.fits'

    # If the machine does not have 8 cores, it will only use up to what is
    # available. We save the catalog to make sure that code runs but we do not
    # compare catalog.
    WfcWrapper(inputfile, binsize=4, extension=4,
               output_root=f'{rootname}_flc',
               output_dir=tmp_path, max_width=37.5,
               processes=8, execute=True, save_mask=True,
               save_diagnostic=False, save_catalog=True)

    # Compare mask with truth
    fdiff = FITSDiff(actual, desired, rtol=1e-7, atol=0)
    if not fdiff.identical:
        raise AssertionError(fdiff.report())



"""This module contains two functions that provide
means to programmatically query the `ACS/WFC
Focus-Diverse ePSF Generator <https://acspsf.stsci.edu>`_ API.

1. :func:`psf_retriever` provides the user with the ability
   to perform downloads for a single image rootname, whereas
2. :func:`multi_psf_retriever` provides the ability to do so
   for multiple rootnames.

In all cases, the focus-diverse ePSFs will be
downloaded to a location specified by the user.

Additionally, this module also contains a function, :func:`interp_epsf`, that allows
users to interpolate the provided ePSF arrays to any arbitrary pixel
coordinates, downsample the ePSF into detector space, or apply subpixel phase shifts.

We strongly recommend you read
`ACS ISR 2018-08 <https://www.stsci.edu/files/live/sites/www/files/home/hst/instrumentation/acs/documentation/instrument-science-reports-isrs/_documents/isr1808.pdf>`_
by A. Bellini et al. and
`ACS ISR 2023-06 <https://www.stsci.edu/files/live/sites/www/files/home/hst/instrumentation/acs/documentation/instrument-science-reports-isrs/_documents/isr2306.pdf>`_
by G. Anand et al. for more details
on these ePSF models and tools before using them.

Examples
--------

Define a folder location for downloaded data:

>>> download_location  = '/Users/username/download_folder/'

Retrieve a single 101x101x90 ePSF FITS file for the given rootname:

>>> from acstools.focus_diverse_epsfs import psf_retriever
>>> retrieved_download = psf_retriever('jds408jsq', download_location)

Retrieve ePSFs based on FLC images specified in an external text file
(with one rootname per line):

>>> from acstools.focus_diverse_epsfs import multi_psf_retriever
>>> retrieved_downloads = multi_psf_retriever('input_ipsoots.txt', download_location)

Retrieve ePSFs using rootnames from Proposal ID 13376 obtained from
`astroquery <https://astroquery.readthedocs.io/>`_:

>>> from astroquery.mast import Observations
>>> obsTable = Observations.query_criteria(
...     obs_collection = 'HST', proposal_id="13376", instrument_name = "ACS/WFC",
...     provenance_name = "CALACS")
>>> dataProducts = Observations.get_product_list(obsTable)
>>> dataProducts = dataProducts[
...     (dataProducts['productSubGroupDescription'] == 'FLC') &
...     (dataProducts['type'] == 'S')]
>>> obs_rootnames = list(dataProducts['obs_id'])
>>> retrieved_downloads = multi_psf_retriever(obs_rootnames, download_location)

Read the retrieved ePSFs from file into NumPy array:

>>> from astropy.io import fits
>>> ePSFs = fits.getdata(retrieved_download)

Interpolate a given ePSF to the given pixel location (0-indexed):

>>> from acstools.focus_diverse_epsfs import interp_epsf
>>> x = 2000  # near the middle of the detector along the X-axis
>>> y = 2000  # near the top of the WFC1 chip
>>> chip = "WFC1" # must specify we are interested in WFC1
>>> interpolated_epsf = interp_epsf(ePSFs, x, y, chip)

Similar to above but obtain the ePSF in detector space (instead of 4x supersampling):

>>> interpolated_epsf = interp_epsf(ePSFs, x, y, chip, pixel_space=True)

Similar to above but obtain the ePSF in detector space and with subpixel offsets:

>>> interpolated_epsf = interp_epsf(
...     ePSFs, x, y, chip, pixel_space=True, subpixel_x=0.77, subpixel_y=0.33)

"""  # noqa: E501
import logging
import os
import re
from pathlib import Path
from urllib.request import urlopen, urlretrieve

import numpy as np
import requests
from requests.auth import HTTPBasicAuth

__taskname__ = "focus_diverse_epsfs"
__author__ = "Gagandeep Anand, Yotam Cohen"
__version__ = "1.0"
__vdate__ = "02-Oct-2023"
__all__ = ["psf_retriever", "multi_psf_retriever", "interp_epsf"]

# Ex: "attachment; filename=jds408jsq-F606W/SM4/STDPBF_ACSWFC_F606W_SM4_F11.2.fits"
_re_patt = re.compile(r"attachment; filename=(\w{9})-\w*\/\w*\/(.*)")

# Initialize the logger
logging.basicConfig()
LOG = logging.getLogger(f'{__taskname__}.Query')
LOG.setLevel(logging.INFO)


def _validate_acs_ipsoot(line):
    return isinstance(line, str) and len(line) == 9 and line.startswith("j")


def psf_retriever(ipsoot, download_location, timeout=60):
    """Function to query API on AWS API Gateway for the ePSF FITS file that
    corresponds to a given image rootname.

    .. warning::

        There is no caching. Please check that you do not have the file
        already downloaded before running this function again.
        Any existing local file with the same name might be overwritten.

    Parameters
    ----------
    ipsoot : str
        Image rootname in the form of IPPPSSOOT.

    download_location : str
        Directory name where the file will be downloaded to.
        It must exist and you must have write permission to it.

    timeout : float
        Seconds before query timeout.

    Returns
    -------
    desired_filename : str or `None`
        The downloaded ePSF FITS file, if successful.

    """
    if not _validate_acs_ipsoot(ipsoot):
        LOG.error("Invalid ACS IPSOOT: %s" % ipsoot)
        return

    if not os.path.isdir(download_location):
        LOG.error("Invalid download location")
        return

    # provide api URL and public key and ID
    api_url = (
        'https://8cclxcxse4.execute-api.us-east-1.amazonaws.com/main/psf-server-ops/')
    api_id = 'iwx1prnqog'
    api_key = 'T1fU5vycfM9KDlSLJoGBU6t0pzS0vjHKaSqmT6gU'

    # stitch together credentials
    auth = HTTPBasicAuth(api_id, api_key)

    # send up post request with ipsoot event
    myobj = {'ipsoot': ipsoot}
    result = requests.post(api_url, json=myobj, auth=auth, timeout=timeout)

    if not result.ok:
        LOG.error("Query failed: %d %s" % (result.status_code, result.reason))
        return

    # grab url from result
    url = result.text[1:-1]
    if not url.startswith("http"):
        LOG.error("URL is not HTTP.")
        return
    with urlopen(url) as remotefile:  # nosec (already checked above)
        # determine readable name for file
        content_disposition = remotefile.info()['Content-Disposition']

    m = _re_patt.match(content_disposition)
    if not m:
        LOG.error("Query failed: No filename in Content-Disposition")
        return

    try:
        desired_filename = os.path.join(download_location, f"{m[1]}-{m[2]}")
    except Exception:
        LOG.error("Query failed: Error extracting filename from Content-Disposition")
        return

    # download file
    urlretrieve(url, filename=desired_filename)  # nosec (already checked above)

    return desired_filename


def multi_psf_retriever(input_list, download_location, num_workers=8):
    """Function to batch query the API on AWS API Gateway for multiple ePSFs
    simultaneously.

    .. note::

        This function requires an optional ``dask`` dependency to be installed.

    Parameters
    ----------
    input_list : list or str
        If a list is given, it must contain a list of rootnames.
        If a string is given, it must be a text file with one rootname per line.

    download_location : str
        Directory name where the file will be downloaded to.
        It must exist and you must have write permission to it.

    num_workers : int
        Max number of workers for ``dask.compute()``.

    Returns
    -------
    results : list of str
        List of downloaded ePSF FITS files.

    """
    import dask
    from dask.diagnostics import ProgressBar

    if not os.path.isdir(download_location):
        LOG.error("Invalid download location")
        return []

    if isinstance(input_list, str):  # Text file with one rootname per line
        orig_lines = Path(input_list).read_text().rsplit()
    elif not isinstance(input_list, (list, tuple)):
        LOG.error("Invalid input list")
        return []
    else:  # List of rootnames
        orig_lines = input_list

    # Should throw out things that are obviously wrong here to minimize network hit.
    lines = [line for line in orig_lines if _validate_acs_ipsoot(line)]
    n_lines = len(lines)

    if n_lines < 1:  # Nothing to do
        return []

    if n_lines == 1:  # No need for multiprocessing
        return [psf_retriever(lines[0], download_location)]

    if n_lines < num_workers:  # No need that many workers
        num_workers = n_lines

    # perform multiprocessing with dask
    dask_results = [dask.delayed(psf_retriever)(line, download_location) for line in lines]

    # run with progress bar
    with ProgressBar():
        results = dask.compute(
            *dask_results, num_workers=num_workers, scheduler='processes')

    return results


def interp_epsf(ePSFs, x, y, chip, pixel_space=False, subpixel_x=0, subpixel_y=0):
    """Function to perform further spatial interpolations given the input ePSF array.
    It uses bi-linear interpolation for the integer pixel shifts, and
    bi-cubic interpolation for any specified subpixel phase shifts.

    This function allows users to interpolate the provided ePSF arrays
    to any arbitrary ``(x, y)`` pixel coordinates. It can be called with
    ``pixel_space=True`` to downsample the ePSF into detector space.

    Subpixel phase shifts can be applied by setting ``subpixel_x`` and ``subpixel_y``
    between 0 and 0.99. Note that a 1 pixel border is removed from the subpixel
    phase shifted ePSF, such that the final dimensions are 23x23. Results from this
    subpixel phase shift routine may differ from other algorithmic
    implementations, typically at the level of <0.5% in the core of the ePSF.

    .. note::

        This function requires an optional ``scipy`` dependency to be installed
        for ``pixel_space=True``.

    .. note::

        This function requires users to specify the WFC chip (WFC1 or WFC2).

    Parameters
    ----------
    ePSFs : numpy.ndarray
        Array with the ePSFs.

    x : int
        X-coordinate (1-indexed) of the desired output ePSF. Please note that the range here is between
        1 and 4096, inclusive. The ePSF grid begins off the detector, at (0,0).

    y : int
        Y-coordinate (1-indexed) of the desired output ePSF. Please note that the range here is between
        1 and 2048, inclusive. The ePSF grid begins off the detector, at (0,0).

    chip : str
        String corresponding to which ACS/WFC detector the user is specifying the coordinates on,
        either "WFC1" or "WFC2".

    pixel_space : bool
        If `True`, downsample the ePSF into detector space.

    subpixel_x, subpixel_y : float
        The desired subpixel coordinate, between 0 and 0.99.
        ``pixel_space`` must be set to `True` to use this option.
        The defaults for both x and y are 0, and are relative to the central pixel of the ePSF.
        By default, ePSFs are centered on the central pixel.

    Returns
    -------
    P : numpy.ndarray or `None`
        An ePSF array with the specified interpolation parameters, if successful.

    See Also
    --------
    psf_retriever, multi_psf_retriever

    """
    valid_wfc_x_pixels = range(1, 4097)
    valid_wfc_y_pixels = range(1, 2049)

    if x not in valid_wfc_x_pixels:
        LOG.error("The X coordinate should be an integer between 1 and 4096.")
        return

    if y not in valid_wfc_y_pixels:
        LOG.error("The Y coordinate should be an integer between 1 and 2048.")
        return

    if subpixel_x < 0 or subpixel_x > 0.99 or subpixel_y < 0.0 or subpixel_y > 0.99:
        LOG.error("Subpixel shifts should be between 0 and 0.99")
        return

    if not isinstance(pixel_space, bool):
        LOG.error("pixel_space should be True or False.")
        return

    if not pixel_space and (subpixel_x != 0 or subpixel_y != 0):
        LOG.error("Please set pixel_space=True to use the subpixel phase offsets.")
        return

    # round subpixel_x and subpixel_y to second decimal
    subpixel_x = round(subpixel_x, 2)
    subpixel_y = round(subpixel_y, 2)

    # give positions of ePSFs based on Andrea and Jay's coordinate system
    # 0,0 is off the detector, but that is where the ePSF definitions begin
    acs_xCoords = np.array([0, 512, 1024, 1536, 2168, 2800, 3192, 3584, 4096])
    acs_yCoords = np.array([0, 512, 1024, 1536, 2048])

    # now, need to find closest four PSFs to any user-defined position

    # first, find coordinates of ePSFs surrounding userCoords
    xLow = acs_xCoords[acs_xCoords <= x].max()
    xHigh = acs_xCoords[acs_xCoords >= x].min()

    yLow = acs_yCoords[acs_yCoords <= y].max()
    yHigh = acs_yCoords[acs_yCoords >= y].min()

    # as well as 2d index of those ePSFs
    xLow_num = list(acs_xCoords).index(xLow)
    xHigh_num = list(acs_xCoords).index(xHigh)

    yLow_num = list(acs_yCoords).index(yLow)
    yHigh_num = list(acs_yCoords).index(yHigh)

    bot_left_psf_index = (xLow_num, yLow_num)
    bot_right_psf_index = (xHigh_num, yLow_num)

    top_left_psf_index = (xLow_num, yHigh_num)
    top_right_psf_index = (xHigh_num, yHigh_num)

    # convert above indices back to single dimensional retrievers
    # for (x = 1000, y = 2000, chip = "WFC2"),
    # should be 28, 29, 37, 38 (assuming index starts at 0)
    bot_left_single_index = bot_left_psf_index[0] + (bot_left_psf_index[1] * 9)
    bot_right_single_index = bot_right_psf_index[0] + (bot_right_psf_index[1] * 9)
    top_left_single_index = top_left_psf_index[0] + (top_left_psf_index[1] * 9)
    top_right_single_index = top_right_psf_index[0] + (top_right_psf_index[1] * 9)

    # do additional conversion if WFC1 (move up 45 indices each)
    if chip == "WFC1":
        bot_left_single_index = bot_left_single_index + 45
        bot_right_single_index = bot_right_single_index + 45
        top_left_single_index = top_left_single_index + 45
        top_right_single_index = top_right_single_index + 45

    # and now do bilinear interpolation
    # using notation from https://www.omnicalculator.com/math/bilinear-interpolation

    # case 1, where user coordinate is equivalent to a known ePSF
    if (bot_left_single_index == bot_right_single_index ==
            top_left_single_index == top_right_single_index):
        c_Q11 = 0.25
        c_Q12 = 0.25
        c_Q21 = 0.25
        c_Q22 = 0.25

    # case 2, where point falls on a y-grid line, so interpolate along x-axis only
    elif bot_left_single_index == top_left_single_index:
        c_Q11 = ((xHigh - x) / (xHigh - xLow))
        c_Q12 = 0
        c_Q21 = (1 - c_Q11)
        c_Q22 = 0

    # case 3, where point falls on an x-grid line, so interpolate along y-axis only
    elif bot_left_single_index == bot_right_single_index:
        c_Q11 = ((yHigh - y) / (yHigh - yLow))
        c_Q12 = (1 - c_Q11)
        c_Q21 = 0
        c_Q22 = 0

    # else do regular bilinear interpolation
    else:
        dxh = (xHigh - x)
        dyh = (yHigh - y)
        dxl = (x - xLow)
        dyl = (y - yLow)
        denom = ((xHigh - xLow) * (yHigh - yLow))
        c_Q11 = (dxh * dyh) / denom
        c_Q21 = (dxl * dyh) / denom
        c_Q12 = (dxh * dyl) / denom
        c_Q22 = (dxl * dyl) / denom

    Q11 = ePSFs[bot_left_single_index]
    Q21 = ePSFs[bot_right_single_index]
    Q12 = ePSFs[top_left_single_index]
    Q22 = ePSFs[top_right_single_index]

    P = Q11 * c_Q11 + Q21 * c_Q21 + Q12 * c_Q12 + Q22 * c_Q22

    # if the user wants the ePSF in pixel space, return that with any subpixel offsets, instead of P.
    if pixel_space:
        from scipy import ndimage

        # shift the ePSF by the specified subpixel amount (assumes no subpixel shift by default)
        P_sub = ndimage.shift(P, (subpixel_y * 4, subpixel_x * 4), order=3)

        # create blank array for downsampled ePSF
        P_sub_down = np.zeros((25, 25))

        # then downsample
        for i in range(25):
            for j in range(25):
                P_sub_down[i, j] = P_sub[4 * i + 2, 4 * j + 2]

        # re-center
        x_int_shift = 0
        y_int_shift = 0

        if subpixel_x >= 0.5:
            x_int_shift = -1

        if subpixel_y >= 0.5:
            y_int_shift = -1

        P_sub_down = ndimage.shift(P_sub_down, (y_int_shift, x_int_shift))

        # remove 1 pixel border
        P_sub_down = P_sub_down[1:-1, 1:-1]

        # and return downsampled version, with any shifts (if specified)
        return P_sub_down

    # otherwise by default return the interpolated 4x supersampled ePSF
    return P



"""
This module contains a class, :class:`Query`, that was implemented to provide
users with means to programmatically query the
`ACS Zeropoints Calculator <https://acszeropoints.stsci.edu>`_ API.
This class works by submitting requests directly to the AWS API, which also
handles the requests from the above web address. It is only valid for ACS
specific instruments (HRC, SBC, or WFC).

The API can be used in two ways by specifying either a
``(date, detector, filter)`` combination or just a ``(date, detector)``
combination. In the first case, the query
will return the zeropoint information for the specific filter and detector at
specified date. In the second case, the query will return the zeropoint
information for all the filters for the desired detector at the specified date.
In either case, the result will be an ``astropy.table.QTable`` where each column
is an ``astropy.units.quantity.Quantity`` object with the appropriate units attached.

Examples
--------

Retrieve the zeropoint information for all of the WFC filters on 2016-04-01:

>>> from acstools import acszpt
>>> q = acszpt.Query(date="2016-04-01", detector="WFC")
>>> zpt_table = q.fetch()
>>> print(zpt_table)
Filter PHOTLAM             PHOTFLAM            STmag  VEGAmag  ABmag
       Angstrom erg / (Angstrom cm2 electron) mag(ST)   mag   mag(AB)
------ -------- ----------------------------- ------- ------- -------
 F435W   4329.9                    3.1858e-19  25.142  25.767  25.652
 F475W   4747.0                     1.845e-19  25.735  26.151  26.045
 F502N   5023.0                    5.2934e-18  22.091  22.367  22.278
 F550M   5581.5                    4.0186e-19   24.89  24.825  24.848
 F555W   5361.0                    1.9798e-19  25.658  25.714  25.704
 F606W   5921.9                    7.8774e-20  26.659  26.402  26.489
 F625W   6311.8                     1.198e-19  26.204  25.728  25.895
 F658N   6584.0                    1.9976e-18  23.149  22.378  22.748
 F660N   6599.4                    5.2219e-18  22.105  21.414    21.7
 F775W   7693.5                    1.0033e-19  26.396   25.26  25.658
 F814W   8045.5                     7.092e-20  26.773  25.506  25.937
F850LP   9031.5                    1.5313e-19  25.937  24.323  24.851
 F892N   8914.8                    1.5105e-18  23.452  21.892  22.394

Retrieve the zeropoint information for the WFC/F435W filter on 2016-04-01:

>>> from acstools import acszpt
>>> q = acszpt.Query(date="2016-04-01", detector="WFC", filt="F435W")
>>> zpt_table = q.fetch()
>>> print(zpt_table)
Filter PHOTLAM             PHOTFLAM            STmag  VEGAmag  ABmag
       Angstrom erg / (Angstrom cm2 electron) mag(ST)   mag   mag(AB)
------ -------- ----------------------------- ------- ------- -------
 F435W   4329.9                    3.1858e-19  25.142  25.767  25.652

Retrieve the zeropoint information for the WFC/F435W filter on multiple dates:

>>> from acstools import acszpt
>>> dates = ['2004-10-13', '2011-04-01', '2014-01-17', '2018-05-23']
>>> queries = []

>>> for date in dates:
...     q = acszpt.Query(date=date, detector='WFC', filt='F435W')
...     zpt_table = q.fetch()
...     # Each object has a zpt_table attribute, so we save the instance
...     queries.append(q)

>>> for q in queries:
...     print(q.date, q.zpt_table['PHOTFLAM'][0], q.zpt_table['STmag'][0])
2004-10-13 3.1111e-19 erg / (Angstrom cm2 electron) 25.168 mag(ST)
2011-04-01 3.1766e-19 erg / (Angstrom cm2 electron) 25.145 mag(ST)
2014-01-17 3.1817e-19 erg / (Angstrom cm2 electron) 25.143 mag(ST)
2018-05-23 3.1897e-19 erg / (Angstrom cm2 electron) 25.141 mag(ST)

>>> type(queries[0].zpt_table['PHOTFLAM'])
astropy.units.quantity.Quantity

"""
import datetime as dt
import json
import logging
import os
import requests

import astropy.units as u
from astropy.table import QTable

__taskname__ = "acszpt"
__author__   = "Gagandeep Anand, Jenna Ryon, Nathan Miles"
__version__  = "2.0"
__vdate__    = "10-Aug-2022"
__all__      = ['ACSZeropointQueryError', 'Query']

# Initialize the logger
logging.basicConfig()
LOG = logging.getLogger(f'{__taskname__}.Query')
LOG.setLevel(logging.INFO)


class ACSZeropointQueryError(Exception):
    """Class used for raising exceptions with API Gateway post requests.
    """
    pass


class Query:
    """Class used to interface with the ACS Zeropoints Calculator API.

    Parameters
    ----------
    date : str
        Input date in the following ISO format, YYYY-MM-DD.

    detector : {'HRC', 'SBC', 'WFC'}
        One of the three channels on ACS: HRC, SBC, or WFC.

    filt : str or `None`, optional
        One of valid filters for the chosen detector. If no filter is supplied,
        all of the filters for the chosen detector will be used:

            * HRC:
                 F220W, F250W, F330W,
                 F344N, F435W, F475W,
                 F502N, F550M, F555W,
                 F606W, F625W, F658N, F660N,
                 F775W, F814W, F850LP, F892N
            * WFC:
                 F435W, F475W,
                 F502N, F550M, F555W,
                 F606W, F625W, F658N, F660N,
                 F775W, F814W, F850LP, F892N
            * SBC:
                 F115LP, F122M, F125LP,
                 F140LP, F150LP, F165LP

    """

    def __init__(self, date, detector, filt=None):

        # Set the attributes
        self._date = date
        self._detector = detector.upper()
        if filt is None:
            self._filt = filt
        else:
            self._filt = filt.upper()

        # define valid detectors and filter combinations
        self._valid_detectors = ('HRC', 'SBC', 'WFC')
        self.valid_filters = {
            'WFC': ['F435W', 'F475W', 'F502N', 'F550M',
                    'F555W', 'F606W', 'F625W', 'F658N',
                    'F660N', 'F775W', 'F814W', 'F850LP', 'F892N'],

            'HRC': ['F220W', 'F250W', 'F330W', 'F344N',
                    'F435W', 'F475W', 'F502N', 'F550M',
                    'F555W', 'F606W', 'F625W', 'F658N',
                    'F660N', 'F775W', 'F814W', 'F850LP', 'F892N'],

            'SBC': ['F115LP', 'F122M', 'F125LP',
                    'F140LP', 'F150LP', 'F165LP']
        }

        self._zpt_table = None
        self._warnings = []

        # ACS Launch Date
        self._acs_installation_date = dt.datetime(2002, 3, 7)
        # end of data table in AWS S3 bucket
        self._end_table_date = dt.datetime(2029, 12, 31)
        self._msg_div = '-' * 79
        self._response = None
        self._failed = False

    @property
    def date(self):
        """The user supplied date. (str)"""
        return self._date

    @property
    def detector(self):
        """The user supplied detector. (str)"""
        return self._detector

    @property
    def filt(self):
        """The user supplied filter, if one was given. (str or `None`)"""
        return self._filt

    @property
    def zpt_table(self):
        """The results returned by the ACS Zeropoint Calculator. (`astropy.table.QTable`)"""
        return self._zpt_table

    def _check_inputs(self):
        """Check the inputs to ensure they are valid.

        Returns
        -------
        status : bool
            True if all inputs are valid, False if one is not.
        """

        valid_detector = True
        valid_filter = True
        valid_date = True

        # Determine the submitted detector is valid
        if self.detector not in self._valid_detectors:
            msg = (f'{self.detector} is not a valid detector option.\n'
                   'Please choose one of the following:\n'
                   f'{os.linesep.join(self._valid_detectors)}\n'
                   f'{self._msg_div}')
            LOG.error(msg)
            valid_detector = False

        # Determine if the submitted filter is valid
        if (self.filt is not None and valid_detector and
                self.filt not in self.valid_filters[self.detector]):
            msg = (f'{self.filt} is not a valid filter for {self.detector}\n'
                   'Please choose one of the following:\n'
                   f'{os.linesep.join(self.valid_filters[self.detector])}\n'
                   f'{self._msg_div}')
            LOG.error(msg)
            valid_filter = False

        # Determine if the submitted date is valid
        date_check = self._check_date()
        if date_check is not None:
            LOG.error(f'{date_check}\n{self._msg_div}')
            valid_date = False

        if not valid_detector or not valid_filter or not valid_date:
            return False

        return True

    def _check_date(self, fmt='%Y-%m-%d'):
        """For determining if the input date is valid.

        Parameters
        ----------
        fmt : str
            The format of the date string. The default is ``%Y-%m-%d``, which
            corresponds to ``YYYY-MM-DD``.

        Returns
        -------
        status : str or `None`
            If the date is valid, returns `None`. If the date is invalid,
            returns a message explaining the issue.
        """

        result = None
        try:
            dt_obj = dt.datetime.strptime(self.date, fmt)
        except ValueError:
            result = f'{self.date} does not match YYYY-MM-DD format'
        else:
            if dt_obj < self._acs_installation_date:
                result = ('The observation date cannot occur '
                          'before ACS was installed '
                          f'({self._acs_installation_date.strftime(fmt)})')
            if dt_obj > self._end_table_date:
                result = ('The observation date cannot be '
                          'after 2029 '
                          f'({self._end_table_date.strftime(fmt)})')
        finally:
            return result

    def fetch(self):
        """Function to query API on AWS APIGateway for zeropoints for single or all
        filters of a given ACS detector (HRC, SBC, or WFC) on a specified date.

        Returns
        -------
        table : `astropy.table.QTable` or `None`
            If the request was successful, returns an astropy Qtable; otherwise, `None`.
        """

        # check user input date, detector, and filter
        bool_inputs = self._check_inputs()

        if bool_inputs:

            # select between all filters or a single filter depending on if user has
            # specified a filter, and then generate a request body to send the API
            if self.filt is None:
                # URL to invoke all_filter API on AWS APIGateway
                invokeURL = "https://vtbopx9sf3.execute-api.us-east-1.amazonaws.com/main/all-filter-resource"  # noqa
                # Generate a request body to send the API
                body = '{"date": "%s", "detector": "%s"}' % (self.date, self.detector)
            else:
                # URL to invoke single_filter API on AWS APIGateway
                invokeURL = "https://vtbopx9sf3.execute-api.us-east-1.amazonaws.com/main/single-filter-resource"  # noqa
                # Generate a request body to send the API
                body = '{"date": "%s", "detector": "%s", "filter": "%s"}' % (self.date, self.detector, self.filt)

            # send request to APIGateway with try/except clauses
            help_desk_string = ("\nIf this error persists, please contact the HST Help Desk at \n"
                                "https://stsci.service-now.com/hst")

            try:
                response = requests.post(invokeURL, data=body, timeout=100)

            except requests.exceptions.HTTPError:
                raise ACSZeropointQueryError(f"HTTP Error to AWS API Gateway:{help_desk_string}") from None

            except requests.exceptions.ConnectionError:
                raise ACSZeropointQueryError(f"Error Connecting to AWS API Gateway:{help_desk_string}") from None

            except requests.exceptions.Timeout:
                raise ACSZeropointQueryError(f"Timeout Error to AWS API Gateway:{help_desk_string}") from None

            except requests.exceptions.RequestException:
                raise ACSZeropointQueryError(f"Request Exception Error to AWS API Gateway:{help_desk_string}") from None

            # and store the results
            APIoutput = json.loads(response.text)

            # define appropriate headers for the table
            headers = ['Detector', 'Filter', 'PHOTLAM', 'PHOTFLAM', 'STmag', 'VEGAmag', 'ABmag']

            # store each row of the results
            rows = APIoutput["rows"]

            # generate an astropy QTable from the results
            table = QTable(rows=rows, names=headers,
                           dtype=('S', 'S', 'float64', 'float64', 'float64', 'float64', 'float64'))

            # remove detector column (to match output of previous API version)
            table.remove_column('Detector')

            # set the appropriate units for each column
            table['Filter'].unit = u.dimensionless_unscaled
            table['PHOTLAM'].unit = u.angstrom
            table['PHOTFLAM'].unit = u.erg / (u.cm * u.cm * u.angstrom * u.electron)
            table['STmag'].unit = u.STmag
            table['VEGAmag'].unit = u.mag
            table['ABmag'].unit = u.ABmag

            self._zpt_table = table

        return self._zpt_table



'''
Various helper functions for :ref:`findsat_mrt`.
'''
import logging
import time
import warnings
from multiprocessing import Pool

import numpy as np
from astropy.convolution import convolve, Gaussian2DKernel
from astropy.io import fits
from astropy.modeling import models, fitting
from astropy.nddata import Cutout2D, block_replicate
from astropy.stats import sigma_clip
from astropy.table import Table, vstack
from astropy.utils.exceptions import AstropyUserWarning

# check for scikit-image
try:
    from skimage import transform
    from skimage.transform._warps import warp
    from skimage._shared.utils import convert_to_float
except ImportError as e:
    warnings.warn(f'skimage not installed. MRT calculation will not work: {repr(e)}')  # noqa

# check for scipy
try:
    from scipy import interpolate
except ImportError:
    warnings.warn('scipy not installed. Kernel generation will not work')

# turn plotting off if matplotlib is not available
try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None
    warnings.warn('matplotlib not found, plotting is disabled')

__taskname__ = "utils_findsat_mrt"
__author__ = "David V. Stark"
__version__ = "1.0"
__vdate__ = "10-Feb-2022"
__all__ = ['merge_tables', 'good_indices', 'fit_streak_profile',
           'rotate_image_to_trail', 'filter_sources', 'create_mask', 'rotate',
           'streak_endpoints', 'streak_persistence', 'add_streak', 'rot_sum',
           'rot_med', 'radon', 'create_mrt_line_kernel', 'update_dq']

# Initialize the logger
logging.basicConfig()
LOG = logging.getLogger(f'{__taskname__}')
LOG.setLevel(logging.INFO)


def _round_up_to_odd(f):
    '''
    Rounds number to nearest odd value.

    Parameters
    ----------
    f : float
        Input number.

    Returns
    -------
    result : float
        Rounded number.

    '''
    return np.ceil(f) // 2 * 2 + 1


def merge_tables(tbls, theta_sep=10, rho_sep=10):
    '''
    Merge multiple trail tables together while removing duplicates.

    Parameters
    ----------
    tbls : list
        List of `~astropy.table.Table` to compare.
    theta_sep : float, optional
        Minimum separation in ``xcentroid`` between catalog entries. Note that
        ``xcentroid`` is the pixel number in the ``theta`` direction. The
        default is 10.
    rho_sep : float, optional
        Minimum separation in ``ycentroid`` between catalog entries. Note that
        ``ycentroid`` is the pixel number in the ``rho`` direction. The default
        is 10.

    Returns
    -------
    src : `~astropy.table.Table`
        Merged table.

    '''
    n_tbls = len(tbls)
    if n_tbls == 0:
        return
    elif n_tbls == 1:
        return tbls[0]

    src = tbls[0]
    for t in tbls[1:]:
        # check that added sources are not recreations of already detected
        # ones
        keep = np.ones(len(t), dtype=bool)
        for s in src:
            dtheta = np.abs(t['xcentroid'] - s['xcentroid'])
            drho = np.abs(t['ycentroid'] - s['ycentroid'])
            sel = (dtheta < theta_sep) & (drho < rho_sep)
            keep[sel] = False
        # Silence MergeConflictWarning for "date" metadata because the tables
        # were created with different timestamps. The last entry is kept.
        src = vstack([src, t[keep]], metadata_conflicts="silent")
    return src


def good_indices(inds, shape):
    '''
    Ensures indices are within bounds of array.

    Parameters
    ----------
    inds : list of tuples
        A list of tuples of the format of ``[(y_min, y_max), (x_min, x_max)]``
        for a 2D array.
    shape : tuple of int
        Shape of the array under consideration, e.g., ``(ny, nx)``.

    Returns
    -------
    fixed_inds : array of tuples
        Same format as input ``inds``, but truncated to fall inside array
        dimensions.

    '''
    # make sure inds is the right format
    if not isinstance(inds, list):
        inds = [inds]
    if isinstance(shape, int):
        shape = [shape]

    if len(inds) != len(shape):
        raise ValueError("Mismatch between inds and shape lengths.")

    # now check individual indices
    good_inds = []
    for (ind1, ind2), sz in zip(inds, shape):
        new_inds = (np.max([0, ind1]), np.min([sz - 1, ind2]))
        good_inds.append(new_inds)

    return good_inds


def fit_streak_profile(yarr, p0, fit_background=True, plot_streak=False,
                       max_width=None, ax=None, bounds=None):
    '''
    Fits a Gaussian to a 1D cross-section of a trail in an image.

    Parameters
    ----------
    yarr : float
        Values of trail cross-section at each position.
    p0 : list of float or `None`
        Initial guesses for amplitude, mean, and sigma parameters.
        Each guess is allowed to be `None` if unknown.
    fit_background : bool, optional
        Set to fit a polynomial to the background. The default is `True`.
    plot_streak : bool, optional
        Set to plot the resulting fitted profile. The default is `False`.
    max_width : int or `None`, optional
        Maximum allowed width of robust satellite trail. Used to define
        background regions. The default is `None`.
    ax : `matplotlib.axes.Axes` or `None`, optional
        Matplotlib subplot axis of the plot. The default is `None` (no plot).
    bounds : dict or `None`, optional
        Dictionary containing the bounds of fit parameters. Format is
        ``{'parameter': (lower, upper)}``. The default is `None`.

    Returns
    -------
    g : `~astropy.modeling.functional_models.Gaussian1D`
        Best-fit 1D Gaussian model.
    snr : float
        Derived signal-to-noise ratio of feature.
    width : float
        Derived 3-sigma width of feature, where sigma is from the
        Gaussian fit.
    mean_flux : float
        The mean flux measured within +/-sigma from the Gaussian fit.

    '''
    amp0, mean0, stdev0 = p0  # initial guesses for amplitude, mean, stddev.
    # Can be None

    # update mean0 to middle of frame if necessary; other will be updated below
    if mean0 is None:
        mean0 = np.len(yarr) / 2 - 0.5

    xarr = np.arange(len(yarr))  # x pixel array

    # subtract baseline level
    if fit_background:

        # try fitting a polynomial to background. Keep to low order to avoid
        # weirdness occuring. E.g., line
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', AstropyUserWarning)
            fit = fitting.LinearLSQFitter()
            or_fit = fitting.FittingWithOutlierRemoval(fit, sigma_clip,
                                                       niter=3, sigma=3)

        # make sure there are regions to fit on either side of the initial
        # position. If not, use lower order to avoid bad behavior
        sel_low = np.where(np.isfinite(yarr) &
                           (xarr < (mean0 - max_width / 2)))[0]
        sel_high = np.where(np.isfinite(yarr) &
                            (xarr > (mean0 + max_width / 2)))[0]
        if (len(sel_high) == 0) or (len(sel_low) == 0):
            order = 1
        else:
            order = 3

        # final fitting region
        sel = np.concatenate([sel_low, sel_high])

        # now run fitting
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', AstropyUserWarning)
            line_init = models.Polynomial1D(degree=order)
            fitted_line, data_mask = or_fit(line_init, xarr[sel], yarr[sel])

        # fitted background
        yarr_bkg = fitted_line(xarr)
    else:
        # if not fitting, subtract median using outer 50% of data
        ind = int(len(yarr) / 4)  # use to select outer quarter of data
        # estimate baseline level using outer 50% of channels
        yarr_bkg = np.nanmedian(list(yarr)[:ind] + list(yarr)[-ind:])

    yarr = yarr - yarr_bkg

    # fit cross-sectional profile

    # update initial guesses if necessary
    if amp0 is None:
        amp0 = np.nanmax(yarr)
    if stdev0 is None:
        stdev0 = 5.

    if bounds is None:
        bounds = {'amplitude': (0, None)}  # forcing positive amplitude

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', AstropyUserWarning)
        g_init = models.Gaussian1D(amplitude=amp0, mean=mean0, stddev=stdev0,
                                   bounds=bounds)
        fit_g = fitting.DogBoxLSQFitter()
    sel = np.isfinite(yarr)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', AstropyUserWarning)
        g = fit_g(g_init, xarr[sel], yarr[sel])

    # plot if triggered
    if plot_streak and (ax is not None) and (plt is not None):
        ax.plot(xarr, yarr)
        ax.plot(xarr[sel], g(xarr[sel]), color='red', label='Fit', lw=3,
                alpha=0.3)

    # characterize the fit

    # measure peak flux directly -- sometimes gaussian not the best fit and
    # can poorly represent the true amplitude
    n_sigma = 3 * g.stddev
    upper, lower = (g.mean + n_sigma, g.mean - n_sigma)
    sel = (xarr > lower) & (xarr < upper)
    if np.sum(sel) == 0:
        peak = 0
    else:
        peak = np.nanmax(yarr[sel])

    # measure mean flux within the standard deviation
    sel = (xarr > (g.mean - g.stddev)) & (xarr < (g.mean + g.stddev))
    if np.sum(sel) == 0:
        mean_flux = 0
    else:
        mean_flux = np.nanmean(yarr[sel])

    # measure noise outside the profile. Using median abs dev to avoid outliers
    sel = (xarr < lower) | (xarr > upper)
    if np.sum(sel) == 0:
        noise = 0
    else:
        noise = np.nanmedian(np.abs(yarr[sel] -
                                    np.nanmedian(yarr[sel]))) / 0.67449

    # use amplitude and noise to re-estimate snr
    snr = (peak - noise) / noise

    LOG.info('amplitude of feature: {}\n'
             'baseline noise: {}\n'
             'snr of feature = {}'.format(peak, noise, snr))

    # use fit parameters to determine the 3-sigma width
    width = upper - lower
    LOG.info('width of feature = {}'.format(width))

    return g, snr, width, mean_flux


def rotate_image_to_trail(image, endpoints):
    '''
    Rotates an image so a given trail runs in the horizontal direction.

    Parameters
    ----------
    image : ndarray
        Image to be rotated.
    endpoints : list of tuples
        List containing the endpoints of the trail. Format is
        ``[(x0, y0), (x1, y1)]``.

    Returns
    -------
    rotated : ndarray
        The rotated image.
    newendpoints : list of tuples
        The endpoints of the rotated trail in the form of
        ``[(rx1, ry1), (rx2, ry2)]``.
    theta : float
        The angle in radians by which the original image was rotated.

    '''
    # extract endpoints
    x1, y1 = endpoints[0]
    x2, y2 = endpoints[1]

    # calculate trail angle
    dy = y2 - y1
    dx = x2 - x1
    theta = np.arctan2(dy, dx)

    # rotate image so trail is horizontal
    rotated = transform.rotate(image, np.degrees(theta), cval=np.nan,
                               resize=True, order=3)

    # now recalculate the new streak endpoints. Need to account for change in
    # image size after rotating

    # get new image center
    xshift = (rotated.shape[1] - image.shape[1]) / 2
    yshift = (rotated.shape[0] - image.shape[0]) / 2
    xx = (image.shape[1] - 1) / 2
    yy = (image.shape[0] - 1) / 2

    # rotate streak endpoints
    rx1, ry1 = rotate((xx, yy), (x1, y1), -theta)
    rx1, ry1 = (rx1 + xshift, ry1 + yshift)
    rx2, ry2 = rotate((xx, yy), (x2, y2), -theta)
    rx2, ry2 = (rx2 + xshift, ry2 + yshift)

    newendpoints = [(rx1, ry1), (rx2, ry2)]

    return rotated, newendpoints, theta


def filter_sources(image, streak_positions, plot_streak=False, buffer=100,
                   minsnr=5, max_width=75, fit_background=True,
                   min_length=50, check_persistence=True, min_persistence=0.5,
                   persistence_chunk=100, min_persistence_snr=3):
    '''
    Filter trail catalog of likely spurious sources.

    The filtering is based on SNR, trail width, and trail_persistence (over
    what fraction of the trail path can the trail actually be well-detected).

    Parameters
    ----------
    image : ndarray
        The original image containing the trails.
    streak_positions : array
        And array containing the endpoints of each trail being analyzed. The
        format is ``[[(x0, y0), (x1, y1)], [(x0_b, y0_b), (x1_b, y1_b)], ...]``.
    plot_streak : bool, optional
        Set to turn on plotting for each analyzed trail (WARNING: THIS CAN
        GENERATE A LOT OF PLOTS). The default is `False`.
    buffer : int, optional
        Size of cutout region on either side of a trail. The default is 100.
    minsnr : float, optional
        Minimum signal-to-noise ratio for a trail to be considered robust.
        The default is 5.
    max_width : int, optional
        Maximum trail width for it to be considered robust. The default is
        75.
    fit_background : bool, optional
        Set to fit a polynomial to the background on either side of an
        extracted trail. The default is `True`.
    min_length : int, optional
        The minimum allowed length of trail to be analyzed. Taking the median
        along each row of the rotated image to create a 1D profile will also
        ignore any rows with fewer than this number of pixels. The default is
        50.
    check_persistence : bool, optional
        Set to turn on persistence checking. The default is `True`.
    min_persistence : float, optional
        Minimum allowed persistence in order for a trail to be considered
        robust. The default is 0.5.
    persistence_chunk : int, optional
        Minimum size of trail sections when breaking it up to measure
        persistence. The default is 100.
    min_persistence_snr : float, optional
        Minimum required signal-to-noise ratio of a chunk of trail in order to
        have a persistence score of 1. The default is 3.

    Returns
    -------
    properties : `~astropy.table.Table`
        Same as the original table but with new columns:

        * mean trail flux
        * trail width
        * trail SNR
        * trail persistence
        * trail status (see :attr:`acstools.utils_findsat_mrt.TrailFinder.mask_include_status`)

    '''  # noqa
    # we'll save the newly measured parameters in the following arrays
    n_streaks = len(streak_positions)
    widths = np.zeros(n_streaks)
    snrs = np.zeros(n_streaks)
    status = np.zeros(n_streaks, dtype=int)
    persistence = np.zeros(n_streaks)
    mean_fluxes = np.zeros(n_streaks)

    # cycle through lines and add them to the mask if they pass tests
    for ii, p in enumerate(streak_positions):

        # rotating so trail is horizontal
        rotated, [[rx1, ry1], [rx2, ry2]], theta = rotate_image_to_trail(image,
                                                                         p)

        # update ry1/2 to include buffer region
        ry1_new = np.min([ry1, ry2]) - buffer
        ry2_new = np.max([ry1, ry2]) + buffer  # making sure ry1 lower than ry2
        streak_y_rot = (ry1 + ry2) / 2  # streak position, possible slight
        # difference in ry1/ry2 due to finite angle sampling

        # buffer region could extend off edge of chip. Truncate ry1/ry2 if so
        fixed_inds = good_indices([(ry1_new, ry2_new), (rx1, rx2)],
                                  rotated.shape)
        ry1_trim, ry2_trim = fixed_inds[0]
        rx1_trim, rx2_trim = fixed_inds[1]

        # find distance of streak from current bottom of the cutout
        dy_streak = streak_y_rot - ry1_trim

        # extract final cutout
        subregion = rotated[int(ry1_trim):int(ry2_trim),
                            int(rx1_trim):int(rx2_trim)]

        # make 1D profile of trail (looking down its axis) by taking a median
        # of all pixels in each row
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore',
                                    message='All-NaN slice encountered')
            medarr = np.nanmedian(subregion, axis=1)

        # get number of pixels being considered at each point; remove those
        # that are too small such that median unreliable
        narr = np.sum(np.isfinite(subregion), axis=1)
        medarr[narr < min_length] = np.nan

        # set up plot
        use_ax = None

        if plot_streak and (plt is not None):
            fig, (ax1, ax2) = plt.subplots(1, 2)
            mad = np.nanmedian(np.abs(subregion))
            ax2.imshow(subregion, vmin=-mad, vmax=5*mad, origin='lower')
            ax2.plot([0, subregion.shape[1] - 1], [dy_streak, dy_streak], '--',
                     lw=2, color='magenta', alpha=0.4)
            use_ax = ax1

        # measure trail properties
        g, snr, width, mean_flux = fit_streak_profile(medarr,
                                                      (None, dy_streak, 5),
                                                      ax=use_ax,
                                                      max_width=max_width,
                                                      plot_streak=plot_streak)
        snrs[ii] = snr
        widths[ii] = width
        mean_fluxes[ii] = mean_flux

        # plot if triggered
        if use_ax:
            ax1.set_title(f'snr={snr}, width={width}')
            ax1.set_xlabel('position')
            ax1.set_ylabel('brightness')
            fig.tight_layout()

        # update status as needed
        if (snr > minsnr) & (width < max_width):
            status[ii] = 1

            # new step; track persistence (if toggled)
            persistence[ii] = -1
            if check_persistence:
                # approximate maximum number of sections and still have sn>3
                # per section (assuming uniform streak)
                maxchunk = np.floor(snr * snr / min_persistence_snr**2)

                dx = persistence_chunk  # starting dx value
                nchunk = np.floor(subregion.shape[1] / dx)  # number of chunks

                # make sure estimated snr per section is >3
                if nchunk > maxchunk:
                    nchunk = maxchunk
                    dx = np.floor(subregion.shape[1] / nchunk)

                if nchunk < 2:
                    # want at least two chunks
                    nchunk = 2
                    dx = np.floor(subregion.shape[1] / nchunk)

                LOG.info('breaking into {} sections for persistence check\n'
                         'Section size for persistence check: {}'.format(nchunk, dx))  # noqa

                persistence[ii] = streak_persistence(subregion, int(dx),
                                                     g.mean.value,
                                                     g.stddev.value,
                                                     max_width=max_width)
                if persistence[ii] > min_persistence:
                    status[ii] = 2

    properties = Table()
    properties['mean flux'] = mean_fluxes
    properties['width'] = widths
    properties['snr'] = snrs
    properties['persistence'] = persistence
    properties['status'] = status
    return properties


def create_mask(image, trail_id, endpoints, widths):
    '''
    Creates an image mask given a set of trail endpoints and widths.

    Parameters
    ----------
    image : ndarray
        Input image to be masked.
    trail_id : array of int
        ID numbers for each trail. These should be unique.
    endpoints : array
        An array containing the endpoints of each trail to contribute to the
        mask. See :func:`filter_sources` for format.
    widths : array
        Widths for each trail.

    Returns
    -------
    segment : ndarray
        A segmentation image where each pixel that is part of a masked
        satellite trail is given a value of the corresponding trail ID.
    mask : ndarray
        Boolean mask (1=masked, 0=not masked).

    '''
    segment = np.zeros(image.shape, dtype=int)

    # cycle through trail endpoints/widths
    for t, e, w in zip(trail_id, endpoints, widths):

        # rotate so trail is horizontal
        rotated, [[rx1, ry1], [rx2, ry2]], theta = rotate_image_to_trail(image,
                                                                         e)

        # create submask using known width
        ry = (ry1 + ry2) / 2.  # take avg, although they should be ~the same
        # max function used to ensure it stays within bounds
        mask_y1 = np.maximum(0, np.floor(ry - w / 2)).astype(int)
        # min function used to ensure it stays within bounds
        mask_y2 = np.minimum(rotated.shape[0] - 1,
                             np.ceil(ry + w / 2)).astype(int)

        mask_x1 = np.maximum(0, np.floor(rx1)).astype(int)
        mask_x2 = np.minimum(rotated.shape[1] - 1, np.ceil(rx2)).astype(int)

        # use these indices to make submask
        submask_rot = np.zeros_like(rotated)
        submask_rot[mask_y1:mask_y2, mask_x1:mask_x2] = t

        # unrotate
        submask_unrot = transform.rotate(submask_rot, -np.degrees(theta),
                                         resize=True, order=1)
        submask_unrot[submask_unrot > 0] = t
        ix0 = (submask_unrot.shape[1] - image.shape[1]) / 2
        ix1 = image.shape[1] + ix0
        iy0 = (submask_unrot.shape[0] - image.shape[0]) / 2
        iy1 = image.shape[0] + iy0
        ix0, ix1, iy0, iy1 = list(map(int, [ix0, ix1, iy0, iy1]))
        subsegment = submask_unrot[iy0:iy1, ix0:ix1]

        # add this to the existing segentation map
        segment = np.maximum(segment, subsegment)

    mask = segment > 0

    return segment, mask


def rotate(origin, point, angle):
    '''
    Rotate a point counterclockwise by a given angle around a given origin.

    Parameters
    ----------
    origin : tuple of float
        ``(x, y)`` coordinates of the origin about which rotation is centered.
    point : tuple, float
        ``(x, y)`` coordinates (in the unrotated reference frame) of the point
        whose new coordinates (in the rotated reference frame) will be
        calculated.
    angle : float
        Angle of rotation in radians.

    Returns
    -------
    qx : float
        Resulting x coordinate in the rotated reference frame.
    qy : float
        Resulting y coordinate in the rotated reference frame.

    '''
    ox, oy = origin
    px, py = point

    qx = ox + np.cos(angle) * (px - ox) - np.sin(angle) * (py - oy)
    qy = oy + np.sin(angle) * (px - ox) + np.cos(angle) * (py - oy)
    return qx, qy


def streak_endpoints(rho, theta, sz, plot=False):
    '''
    Get endpoints of a trail using rho and theta coordinates and image size.

    Parameters
    ----------
    rho : float
        Offset coordinate.
    theta : float
        Angle coordinate in degrees.
    sz : tuple of int
        Size of image in ``(ny, nx)``. Streak endpoints will be truncated at
        the images edges.

    Returns
    -------
    pos : array
        Array with form ``[p0, p1]`` where each ``p`` is ``(x, y)``, the x and
        y coordinates of each end.

    '''
    x0 = sz[1] / 2 - 0.5
    y0 = sz[0] / 2 - 0.5
    sz1_m1 = sz[1] - 1
    sz0_m1 = sz[0] - 1

    theta_rad = np.radians(theta)
    sin_theta = np.sin(theta_rad)
    cos_theta = np.cos(theta_rad)
    dy = rho * sin_theta
    dx = rho * cos_theta
    x1 = x0 + dx
    y1 = y0 + dy

    slope = sin_theta / cos_theta

    # get perpendicular slope/intercept (this is the line that represents what
    # the streak looks like)
    slope_int = -1. / slope

    # special values if the slope is infinity (when streak fully verticle)
    if ~np.isfinite(slope_int):
        no_slope = True
        zero_slope = False
        yi = 0
        yf = sz0_m1
        xi = xf = x0 + rho
    # special values if the slope is close to zero (streak horizontal)
    elif np.allclose(slope_int, 0):
        zero_slope = True
        no_slope = False
        yi = yf = y0 + rho
        xi = 0
        xf = sz1_m1
    else:
        no_slope = False
        zero_slope = False
        b_int = y1 - slope_int * x1

        # get coordinates of streak. Start with x = 0 and max value
        xi, xf = (0, sz1_m1)
        yi, yf = slope_int * np.array([xi, xf]) + b_int

        # update coordinates if yi goes out of bounds
        if yi < 0:
            yi = 0
        elif yi >= sz[0]:
            yi = sz0_m1
        xi = (yi - b_int) / slope_int

        # same things for yf
        if yf < 0:
            yf = 0
        elif yf >= sz[0]:
            yf = sz0_m1
        xf = (yf - b_int) / slope_int

    if plot and (plt is not None):
        fig, ax = plt.subplots(figsize=(10, 10))

        ax.plot([0, sz1_m1], [y0, y0], '--', color='gray')
        ax.plot([x0, x0], [0, sz0_m1], '--', color='gray')

        ax.plot([x0, x1], [y0, y1], color='red')
        ax.set_xlim(0, sz1_m1)
        ax.set_ylim(0, sz0_m1)

        # Define xarr, yarr manually if slope_int is inf
        if no_slope:
            xarr = [xi, xi]
            yarr = [0, sz0_m1]
        elif zero_slope:
            xarr = [xi, xf]
            yarr = [yi, yf]
        else:
            xarr = np.array([0, sz1_m1])
            yarr = slope_int * xarr + b_int

        ax.plot(xarr, yarr, color='blue')
        ax.scatter([xi, xf], [yi, yf], s=100, color='magenta')

    return [(xi, yi), (xf, yf)]  # p0, p1


def streak_persistence(cutout, dx, streak_y0, streak_stdev, max_width=None,
                       plot_streak=False):
    '''
    Measure streak persistence across image.

    The persistence is measured by breaking the trail into even chunks and
    trying to fit a Gaussian to a 1D cross-section from each chunk. Successful
    fits contribute to the persistence score.

    Parameters
    ----------
    cutout : ndarray
        Cutout of the trail. This should be the same cutout generated in
        :func:`filter_sources` so that the trail is aligned horizontally across
        the cutout.
    dx : int
        Size of each chunk to be analyzed.
    streak_y0 : float
        y-coordinate of the trail derived from the analysis of its global
        cross-section.
    streak_stdev : float
        Gaussian sigma value determined from the analysis of its global
        cross-section.
    max_width : int, optional
        Maximum allowed width for a trail to be considered robust. This
        should be the same as whatever was used in :func:`filter_sources`. The
        default is `None`.
    plot_streak : bool, optional
        Set to turn on plotting. WARNING: THIS CAN GENERATE MANY PLOTS
        DEPENDING ON HOW MANY SECTIONS A TRAIL IS BROKEN INTO. The default
        is `False`.

    Returns
    -------
    pscore : float
        Persistence score of the input trail.

    '''
    # starting guess
    guess = (None, streak_y0, streak_stdev)

    # bounds of fit
    bounds = {'amplitude': (0, None), 'mean': (streak_y0, streak_y0 + 25)}

    snr_arr = []
    width_arr = []
    mean_arr = []
    persist = []

    # step through and fit as a function of x
    nsteps = int(np.floor(cutout.shape[1] / dx))
    for ii in range(nsteps):

        LOG.info('Checking persistence, step {} of {}'.format(ii + 1, nsteps))

        ind0 = ii * dx
        ind1 = (ii + 1) * dx

        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore',
                                    message='All-NaN slice encountered')
            chunk = np.nanmedian(cutout[:, ind0:ind1], axis=1)

        if plot_streak and (plt is not None):
            fig, ax = plt.subplots()
        else:
            ax = None

        # fit properties of chunk
        g, snr, width, mean_flux = fit_streak_profile(chunk, guess, ax=ax,
                                                      max_width=max_width,
                                                      plot_streak=plot_streak,
                                                      bounds=bounds)
        if ax:
            ax.set_title(f'{ii}-snr={snr},width={width},mean={g.mean.value}')

        LOG.info('Chunk SNR, width, mean: {}, {}, {}'.format(snr, width,
                                                             g.mean.value))
        snr_arr.append(snr)
        width_arr.append(width)
        mean_arr.append(g.mean.value)

        # assess if successful
        success = (snr > 3) & np.isfinite(snr)
        if ii > 0:
            jj = ii - 1
            success = success & (np.abs(mean_arr[ii] - mean_arr[jj]) < width_arr[jj])  # noqa
            if max_width is not None:
                success = success & (width <= max_width)
        if success:
            persist.append(success)
            # update guess
            guess = (g.amplitude.value, g.mean.value, g.stddev.value)
            bounds = {'amplitude': (0, None), 'mean': (g.mean.value - width,
                                                       g.mean.value + width)}
        else:
            persist.append(False)
            LOG.info('fit failed, will not update guesses')

    # save the persistence score
    pscore = np.sum(persist) / len(persist)

    LOG.info('Number of sections analyzed: {}\n'
             'Number of sections that passed: {}\n'
             'persistance score: {}'.format(len(persist), np.sum(persist),
                                            pscore))

    return pscore


def add_streak(image, width, value, rho=None, theta=None, endpoints=None,
               psf_sigma=None):
    '''
    Generates a (optionally) PSF-convolved model trail and add it to an image.

    Parameters
    ----------
    image : ndarray
        The input image onto which the trail should be added.
    width : float
        Width of the trail. Note that as of now this function does not super-
        sample the image in order to precisely model the trail width.
    value : float
        Uniform intensity within the trail prior to convolving with a PSF.
    rho : float or `None`, optional
        Trail offset from center. The default is `None`.
    theta : float or `None`, optional
        Trail angle in degrees with respect to x-axis. The default is `None`.
    endpoints : array of tuples or `None`, optional
        Endpoints of the trail. Format is ``[(x0, y0), (x1, y1)]``. The default
        is `None`.
    psf_sigma : float or `None`, optional
        Sigma of Gaussian model PSF to be convolved with trail image. The
        default is `None`, in which case no convolution is performed.

    Returns
    -------
    image : ndarray
        The input image with the streak added on.

    '''
    # make sure either rho, theta, or endpoints are set
    if endpoints is None:
        if (rho is not None) and (theta is not None):
            endpoints = streak_endpoints(rho, theta, image.shape)
            LOG.info('calculated endpoints: {}'.format(endpoints))
        else:
            LOG.error('Both rho and theta must be set but got rho={} and '
                      'theta={}'.format(rho, theta))
            return image
    elif (rho is not None) or (theta is not None):
        LOG.warning('rho/theta and endpoints set, defaulting to using '
                    'endpoints')

    # rotate image so trail horizontal
    rotated, newendpoints, rot_theta = rotate_image_to_trail(image, endpoints)

    # add trail
    x1, y1 = newendpoints[0]
    x2, y2 = newendpoints[1]

    trail_y = (y1 + y2) / 2
    trail_image_rot = np.zeros_like(rotated)
    # find edge of trail, avoid going beyond image boundaries
    half_width = width / 2
    max_ind = int(np.minimum(trail_y + half_width + 1, rotated.shape[0]))
    min_ind = int(np.maximum(trail_y - half_width, 0))
    trail_image_rot[min_ind:max_ind, :] = value

    # convolve with PSF if set
    if psf_sigma is not None:
        kernel = Gaussian2DKernel(x_stddev=psf_sigma, y_stddev=psf_sigma)
        trail_image_rot = convolve(trail_image_rot, kernel, boundary='extend')

    # now rotate back
    trail_image = transform.rotate(trail_image_rot, -np.degrees(rot_theta),
                                   resize=True, order=1)
    ix0 = (trail_image.shape[1] - image.shape[1]) / 2
    ix1 = image.shape[1] + ix0
    iy0 = (trail_image.shape[0] - image.shape[0]) / 2
    iy1 = image.shape[0] + iy0
    ix0, ix1, iy0, iy1 = list(map(int, [ix0, ix1, iy0, iy1]))

    trail_image = trail_image[iy0:iy1, ix0:ix1]

    return image + trail_image


def _rot(image, angle, return_length, func):
    center = image.shape[0] // 2
    cos_a, sin_a = np.cos(angle), np.sin(angle)
    R = np.array([[cos_a, sin_a, -center * (cos_a + sin_a - 1)],
                  [-sin_a, cos_a, -center * (cos_a - sin_a - 1)],
                  [0, 0, 1]])

    # rotate image
    rotated = warp(image, R, clip=False, cval=np.nan)

    # take func along each column
    with warnings.catch_warnings():
        # suppressing this warning as it's inconsequential and expected
        warnings.filterwarnings(action='ignore',
                                message='All-NaN slice encountered')
        medarr = func(rotated, axis=0)

    # get length along each column
    if return_length:
        length = np.sum(np.isfinite(rotated), axis=0)
        return [medarr, length]
    return medarr


def rot_sum(image, angle, return_length):
    '''
    Rotates and image by a designated angle and sums the values in each column.

    Parameters
    ----------
    image : ndarray
        Image to be rotated and summed.
    angle : float
        Angle by which to rotate the input image (radians).
    return_length : bool
        Set to the return the number of valid pixels in each column.

    Returns
    -------
    medarr : ndarray
        The resulting sum in each column.
    length : ndarray, optional
        The number of valid pixels in each column.
        This is only returned if ``return_length`` is `True`.

    '''
    return _rot(image, angle, return_length, np.nansum)


def rot_med(image, angle, return_length):
    '''
    Rotates and image by a designated angle and take a median in each column.

    Parameters
    ----------
    image : ndarray
        Image to be rotated and medianed.
    angle : float
        Angle by which to rotate the input image (radians).
    return_length : bool
        Set to the return the number of valid pixels in each column.

    Returns
    -------
    medarr : ndarray
        The resulting median in each column.
    length : ndarray, optional
        The number of valid pixels in each column.
        This is only returned if ``return_length`` is `True`.

    '''
    return _rot(image, angle, return_length, np.nanmedian)


# TODO: If radon performance is improved upstream, we should just use
#       the version in scikit-image and remove this one. See
#       https://github.com/scikit-image/scikit-image/issues/3118
def radon(image, theta=None, circle=False, *, preserve_range=False,
          fill_value=np.nan, median=True, processes=1, return_length=False,
          print_calc_times=False):
    """
    Calculates the (median) radon transform of an image.

    This routine basically :func:`skimage.transform.radon` but with
    extra options and multiprocessing.
    For further information see [1]_ and [2]_.

    Parameters
    ----------
    image : array-like
        Input image. The rotation axis will be located in the pixel with
        indices ``(image.shape[0] // 2, image.shape[1] // 2)``.
    theta : array-like, optional
        Projection angles (in degrees). If `None`, the value is set to
        ``np.arange(180)``.
    circle : bool, optional
        Assume image is zero outside the inscribed circle, making the
        width of each projection (the first dimension of the sinogram)
        equal to ``min(image.shape)``.
    preserve_range : bool, optional
        Whether to keep the original range of values. Otherwise, the input
        image is converted according to the conventions of ``img_as_float``.
        Also see https://scikit-image.org/docs/dev/user_guide/data_types.html
    fill_value : float, optional
        Value to use for regions where the transform could not be calculated.
        Default is 0.
    median: bool, optional
        Flag to turn on Median Radon Transform instead of standard Radon
        Transform. Default is `True`.
    processes: int, optional
        Number of processes to use when calculating the transform. Default is 1
        (no multi-processing).
    return_length : bool, optional
        Option to return an array giving the length of the data array used to
        calculate the transform at every location. Default is `False`.
    print_calc_times : bool, optional
        Log the run time.

    Returns
    -------
    radon_image : ndarray
        Radon transform (sinogram). The tomography rotation axis will lie
        at the pixel index ``radon_image.shape[0] // 2`` along the 0th
        dimension of ``radon_image``.
    length: ndarray, optional
        Length of data array. This is only returned if ``return_length`` is
        `True`.

    Raises
    ------
    ValueError
        Input image is not 2-D or padded image is not square.

    References
    ----------
    .. [1] AC Kak, M Slaney, "Principles of Computerized Tomographic
           Imaging", IEEE Press 1988.
    .. [2] B.R. Ramesh, N. Srinivasa, K. Rajgopal, "An Algorithm for Computing
           the Discrete Radon Transform With Some Applications", Proceedings of
           the Fourth IEEE Region 10 International Conference, TENCON '89, 1989

    Notes
    -----
    Based on code of Justin K. Romberg
    (https://www.clear.rice.edu/elec431/projects96/DSP/bpanalysis.html)

    """
    total_median_time = 0.
    total_warp_time = 0.

    if median is True:
        LOG.info('Calculating median Radon Transform with {} processes'.format(processes))  # noqa
    else:
        LOG.info('Calculating standard Radon Transform with {} processes'.format(processes))  # noqa
    if image.ndim != 2:
        raise ValueError('The input image must be 2-D')
    if theta is None:
        theta = np.arange(180)

    image = convert_to_float(image, preserve_range)

    if circle:
        shape_min = min(image.shape)
        radius = shape_min // 2
        img_shape = np.array(image.shape)
        coords = np.array(np.ogrid[:image.shape[0], :image.shape[1]],
                          dtype=object)
        dist = ((coords - img_shape // 2) ** 2).sum(0)
        outside_reconstruction_circle = dist > radius ** 2
        if np.any(image[outside_reconstruction_circle]):
            warnings.warn('Radon transform: image must be zero outside the '
                          'reconstruction circle', AstropyUserWarning)
        # Crop image to make it square
        slices = tuple(slice(int(np.ceil(excess / 2)),
                             int(np.ceil(excess / 2) + shape_min))
                       if excess > 0 else slice(None)
                       for excess in (img_shape - shape_min))
        padded_image = image[slices]
    else:
        diagonal = np.sqrt(2) * max(image.shape)
        pad = [int(np.ceil(diagonal - s)) for s in image.shape]
        new_center = [(s + p) // 2 for s, p in zip(image.shape, pad)]
        old_center = [s // 2 for s in image.shape]
        pad_before = [nc - oc for oc, nc in zip(old_center, new_center)]
        pad_width = [(pb, p - pb) for pb, p in zip(pad_before, pad)]
        padded_image = np.pad(image, pad_width, mode='constant',
                              constant_values=fill_value)

    # padded_image is always square
    if padded_image.shape[0] != padded_image.shape[1]:
        raise ValueError('padded_image must be a square')
    center = padded_image.shape[0] // 2
    radon_image = np.zeros((padded_image.shape[0], len(theta)),
                           dtype=image.dtype) + np.nan
    lengths = np.copy(radon_image)

    if processes <= 1:
        for i, angle in enumerate(np.deg2rad(theta)):
            cos_a, sin_a = np.cos(angle), np.sin(angle)
            R = np.array([[cos_a, sin_a, -center * (cos_a + sin_a - 1)],
                          [-sin_a, cos_a, -center * (cos_a - sin_a - 1)],
                          [0, 0, 1]])
            warp_time_0 = time.time()
            rotated = warp(padded_image, R, clip=False)
            warp_time_1 = time.time()
            total_warp_time += (warp_time_1 - warp_time_0)
            if median is False:
                radon_image[:, i] = np.nansum(rotated, axis=0)
            else:
                median_time_0 = time.time()
                radon_image[:, i] = np.nanmedian(rotated, axis=0)
                median_time_1 = time.time()
                total_median_time += (median_time_1 - median_time_0)
            lengths[:, i] = np.sum(np.isfinite(rotated), axis=0)

    else:
        # splitting calculation up among many processes to speed up. Each
        # thread rotates and sums/medians at a specific angle
        p = Pool(processes)
        angles = np.deg2rad(theta)
        images = [padded_image for i in range(len(angles))]
        return_lengths = [True for i in range(len(angles))]
        pairs = list(zip(images, angles, return_lengths))
        if median is False:
            result = p.starmap(rot_sum, pairs)
            result = np.array(result)
            radon_image = result[:, 0, :]
            lengths = result[:, 1, :]
        else:
            result = p.starmap(rot_med, pairs)
            result = np.array(result)
            # print(np.shape(result))  # DEBUG
            radon_image = result[:, 0, :]
            lengths = result[:, 1, :]
        radon_image = radon_image.T
        lengths = lengths.T
        p.close()

    if print_calc_times:
        LOG.info('Total time to warp images = {} seconds'.format(total_warp_time))  # noqa
        LOG.info('Total time to calculate medians = {} seconds'.format(total_median_time))  # noqa

    if return_length is True:
        return radon_image, lengths
    else:
        return radon_image


def update_dq(filename, ext, mask, dqval=16384, verbose=True,
              expand_mask=False):
    """Update the given image and DQ extension with the given
    satellite trails mask and flag.

    This function was moved over from `acstools.satdet`.

    Parameters
    ----------
    filename : str
        FITS image filename to update.
    ext : int, str, or tuple
        DQ extension, as accepted by ``astropy.io.fits``, to update.
    mask : ndarray
        Boolean mask, with `True` marking the satellite trail(s).
        This can be the result(s) from :func:`acstools.satdet.make_mask`.
    dqval : int, optional
        DQ value to use for the trail. Default value of 16384 is
        tailored for ACS/WFC.
    verbose : bool, optional
        Print extra information to the terminal.
    expand_mask : bool, optional
        Allows the mask to be expanded to match the size of the original mask.
        This is relevant for masks that were generated used binned versions of
        the original image.

    """
    with fits.open(filename, mode='update') as pf:
        dqarr = pf[ext].data

        # check that the old mask and new mask are the same size
        same_size = mask.shape == dqarr.shape
        mask_larger = np.any([mask.shape[i] > dqarr.shape[i]
                              for i in range(mask.ndim)])

        # provide informative error messages depending on mask/dqarr size
        if not same_size:
            LOG.info('Inconsistent mask sizes: \n'
                     'Input mask: {} \n'
                     'DQ array: {}'.format(mask.shape, dqarr.shape))

            if mask_larger:
                LOG.error('Mask is larger than the DQ array in one or more '
                          'dimensions.\nThis routine cannot currently rescale '
                          'masks that are larger than the origin input image')
                return

            elif not mask_larger and not expand_mask:
                LOG.error('Cannot proceed due to size mismatch.\n'
                          'Set expand_mask=True if the mask was generated '
                          'with binned data and needs to be enlarged to the '
                          'original size.')
                return
            else:
                LOG.info('Enlarging mask to original size.')

                # Determine factor by which to scale-up image. Scales should
                # be the same for the x and y dimensions, after we round down
                # to the nearest integer in case there are extra rows/columns
                # that do not divide evenly. These extra rows/columns will be
                # added back at the end. This approach is the reverse of
                # astropy.nddata.block_reduce, which trims extra rows/cols
                # first.

                factor_x = np.floor(dqarr.shape[1]/mask.shape[1]).astype(int)
                factor_y = np.floor(dqarr.shape[0]/mask.shape[0]).astype(int)
                if factor_x != factor_y:

                    mask_aspect = mask.shape[1]/mask.shape[0]
                    dq_aspect = dqarr.shape[1]/dqarr.shape[0]

                    raise ValueError('Mask and DQ array axis aspect ratios are'
                                     ' too different\n'
                                     'Mask dimensions (x,y) = {}\n'
                                     'Mask aspect ratio (x/y) = {}\n'
                                     'DQ array dimensions (y,x) = {}\n'
                                     'DQ array aspect ratio (x/y) = {}'
                                     .format(mask.shape, mask_aspect,
                                             dqarr.shape, dq_aspect))

                else:
                    factor = factor_x

                # assess whether there will be extra rows/columns after scaling
                extra_cols = dqarr.shape[1] - mask.shape[1]*factor
                extra_rows = dqarr.shape[0] - mask.shape[0]*factor

                mask = block_replicate(mask, factor,
                                       conserve_sum=False)

                if extra_rows > 0:
                    LOG.info('Duplicating end rows to match orig size')
                    while mask.shape[0] != dqarr.shape[0]:
                        mask = np.vstack([mask, mask[-1, :]])

                if extra_cols > 0:
                    LOG.info('Duplicating end columns to match orig size')
                    while mask.shape[1] != dqarr.shape[1]:
                        mask = np.hstack([mask, mask[:, -1].reshape(mask.shape[0], 1)])  # noqa

        old_mask = (dqval & dqarr) != 0  # Existing flagged trails
        new_mask = mask & ~old_mask  # Only flag previously unflagged trails
        npix_updated = np.count_nonzero(new_mask)

        # Update DQ extension only if necessary
        if npix_updated > 0:
            pf[ext].data[new_mask] += dqval
            pf['PRIMARY'].header.add_history(
                f'{time.ctime()} {__taskname__} v{__version__}({__vdate__})')
            pf['PRIMARY'].header.add_history(
                f'  Updated {npix_updated} px in EXT {ext} with DQ={dqval}')

    if verbose:
        fname = f'{filename}[{ext}]'

        print(f'DQ flag value is {dqval}\n'
              f'Input... flagged NPIX={np.count_nonzero(mask)}\n'
              f'Existing flagged NPIX={np.count_nonzero(old_mask)}\n'
              f'Newly... flagged NPIX={npix_updated}')

        if npix_updated > 0:
            print(f'{fname} updated')
        else:
            print(f'No updates necessary for {fname}')


def create_mrt_line_kernel(width, sigma, outfile=None, shape=(1024, 2048),
                           plot=False, theta=None, processes=1):
    '''
    Creates a model signal MRT signal of a line of specified width.

    The trails can be blurred by a simple Gaussian model PSF. More accurate
    PSF convolution using an input PSF model will be incorporated in the
    future. These kernels are used for detection of real linear signals in the
    MRT of imaging data.

    Parameters
    ----------
    width : int
        Width of the line. Intensity is constant over this width.
    sigma : float
        Gaussian sigma of the PSF. This is NOT FWHM.
    outfile : str or `None`, optional
        Location to save an output fits file of the kernel. The default is
        `None`. If a filename is given and it exists, it will be overwritten.
    sz : tuple of int, optional
        Size of the image on which to place the line. The default is
        ``(1024, 2048)``.
    plot : bool, optional
        Flag to plot the original image, MRT, and kernel cutout.
    theta : array or `None`, optional
        Set of angles in degrees at which to calculate the MRT, default is
        `None`, which defaults to ``np.arange(0, 180, 0.5)``.
    processes: int, optional
        Number of processes to use when calculating MRT. Default is 1
        (no multi-processing).

    Returns
    -------
    kernel : ndarray
        The resulting kernel.

    '''
    # updating some defaults
    if theta is None:
        theta = np.arange(0, 180, 0.5)

    # set up empty image and coordinates
    image = np.zeros(shape)
    y0 = image.shape[0] / 2 - 0.5
    x0 = image.shape[1] / 2 - 0.5
    xarr = np.arange(image.shape[1]) - x0
    yarr = np.arange(image.shape[0]) - y0
    x, y = np.meshgrid(xarr, yarr)

    # add a simple streak across the image.
    image = add_streak(image, width, 1, rho=0, theta=90, psf_sigma=sigma)

    # plot the model image
    ax = None
    if plot and (plt is not None):
        fig, ax = plt.subplots()
        ax.imshow(image, origin='lower')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_title('model image')

    # calculate the RT for this model
    rt = radon(image, circle=False, median=True, fill_value=np.nan,
               processes=processes, return_length=False)

    # plot the MRT
    if ax:
        fig2, ax2 = plt.subplots()
        ax2.imshow(rt, aspect='auto', origin='lower')
        ax2.set_xlabel('angle pixel')
        ax2.set_ylabel('offset pixel')
        ax2.set_title('Model MRT')

    # find the center of the signal by summing along each direction and finding
    # the max.
    rt_rho = np.nansum(rt, axis=1)
    rt_theta = np.nansum(rt, axis=0)
    rho0 = np.nanargmax(rt_rho)
    theta0 = np.nanargmax(rt_theta)

    # plot the 1D slices
    if ax:
        fig3, [ax3a, ax3b] = plt.subplots(1, 2)
        ax3a.plot(rt_theta, '.')
        ax3b.plot(rt_rho, '.')

        ax3a.plot([theta0, theta0], [0, 8])
        ax3b.plot([rho0, rho0], [0, 1])
        ax3a.set_xlim(theta0 - 5, theta0 + 5)
        ax3b.set_xlim(rho0 - 10, rho0 + 10)

        ax3a.set_xlabel('theta pixel')
        ax3a.set_ylabel('summed value')

        ax3b.set_xlabel('rho pixel')
        ax3b.set_ylabel('summed value')

    # may need to refine center coords. Run a Gaussian fit to see if necessary
    g_init = models.Gaussian1D(mean=rho0)
    fit_g = fitting.LevMarLSQFitter()
    g = fit_g(g_init, np.arange(len(rt_rho)), rt_rho)
    rho0_gfit = g.mean.value

    g_init = models.Gaussian1D(mean=theta0)
    fit_g = fitting.LevMarLSQFitter()
    g = fit_g(g_init, np.arange(len(rt_theta)), rt_theta)
    theta0_gfit = g.mean.value

    # see if any difference between simple location of max pixel vs. gauss fit
    theta_shift = theta0_gfit - theta0
    rho_shift = rho0_gfit - rho0

    # get initial cutout
    position = (theta0, rho0)
    dtheta = 3
    drho = np.ceil(width / 2 + 3 * sigma)

    size = (_round_up_to_odd(2 * drho), _round_up_to_odd(2 * dtheta))
    cutout = Cutout2D(rt, position, size)

    # inteprolate onto new grid if necessary. Need to generate cutout first.
    # The rt can be too big otherwise
    do_interp = (np.abs(rho_shift) > 0.1) or (np.abs(theta_shift) > 0.1)
    if do_interp:
        LOG.info('Inteprolating onto new grid to center kernel')
        theta_arr = np.arange(cutout.shape[1])
        rho_arr = np.arange(cutout.shape[0])

        new_theta_arr = theta_arr + theta_shift
        new_rho_arr = rho_arr + rho_shift
        new_theta_grid, new_rho_grid = np.meshgrid(new_theta_arr, new_rho_arr)

        f = interpolate.RegularGridInterpolator((rho_arr, theta_arr),
                                                cutout.data,
                                                bounds_error=False,
                                                fill_value=0,
                                                method='cubic')

        # unable to provide 2D grid of points for interpolation. Switching to
        # 1D, and then will switch back
        points = [[r, t] for r, t in zip(np.ravel(new_rho_grid),
                                         np.ravel(new_theta_grid))]

        # return to original shape and replace the old cutout
        cutout = np.reshape(f(points), cutout.data.shape)

    if ax:
        fig4, ax4 = plt.subplots()
        ax4.imshow(cutout.data, origin='lower', aspect='auto')
        ax4.set_title('final kernel')
        ax4.set_xlabel('theta pixel')
        ax4.set_ylabel('rho pixel')

    if outfile is not None:
        fits.writeto(outfile, cutout.data, overwrite=True)

    return cutout.data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module identifies satellite trails in ACS/WFC imaging with the Median
Radon Transform (MRT).

It contains a class called `~acstools.findsat_mrt.TrailFinder` that is used to
identify satellite trails and/or other linear features in astronomical image
data. To accomplish this goal, the MRT is calculated for an image. Point
sources are then extracted from the MRT and filtered to yield a final catalog
of trails, which can then be used to create a mask.

A second class called `~acstools.findsat_mrt.WfcWrapper` is designed explicitly
to make ACS/WFC data easy to process.

This algorithm is found to be roughly 10x more sensitive compared to the
current satellite trail finding code included with acstools, :ref:`satdet`.
However, this approach can struggle with dense fields, while the performance
of :ref:`satdet` in these fields may be more reliable (but this has not yet
been tested).

For further details on this algorithm and tests of its performance, see the
`ISR ACS 2022-8 <https://ui.adsabs.harvard.edu/abs/2022acs..rept....8S/abstract>`_.

Examples
--------

**Example 1:** Identification of trails in an ACS/WFC image,
``j97006j5q_flc.fits`` (the second science and DQ extensions).

To load the data:

>>> import numpy as np
>>> from astropy.io import fits
>>> from acstools.findsat_mrt import TrailFinder
>>> file = 'j97006j5q_flc.fits'
>>> with fits.open(file) as h:
>>>     image = h['SCI', 2].data
>>>     dq = h['DQ', 2].data

Mask some bad pixels, remove median background, and rebin the data to speed up
MRT calculation:

>>> from astropy.nddata import bitmask, block_reduce
>>> mask = bitmask.bitfield_to_boolean_mask(
...     dq, ignore_flags=[4096, 8192, 16384])
>>> image[mask] = np.nan
>>> image = image - np.nanmedian(image)
>>> image = block_reduce(image, 4, func=np.nansum)

Initialize `~acstools.findsat_mrt.TrailFinder` and run these steps:

>>> s = TrailFinder(image, processes=8)  # initialization
>>> s.run_mrt()                       # calculates MRT
>>> s.find_mrt_sources()              # finds point sources in MRT
>>> s.filter_sources()                # filters sources from MRT
>>> s.make_mask()                     # makes a mask from the identified trails
>>> s.save_output()        # saves the output

The input image, mask, and MRT (with sources overlaid) can be plotted during
this process:

>>> s.plot_mrt(show_sources=True)      # plots MRT with sources overlaid
>>> s.plot_image(overlay_mask=True)    # plots input image with mask overlaid

**Example 2:** Quick run to find satellite trails.

After loading and preprocessing the image (see example above), run
the following:

>>> s = TrailFinder(image, processes=8)  # initialization
>>> s.run_all()                              # runs everything else

**Example 3:** Run identification/masking using the WFC wrapper.

`~acstools.findsat_mrt.WfcWrapper` can automatically do the binning, background
subtraction, and bad pixel flagging:

>>> from acstools.findsat_mrt import WfcWrapper
>>> w = WfcWrapper('jc8m32j5q_flc.fits', preprocess=True, extension=4, binsize=2)

In all other respects, it behaves just like `~acstools.findsat_mrt.TrailFinder`,
so to continue the process:

>>> w.run_mrt()
>>> w.find_sources()
>>> w.filter_sources()
>>> w.make_mask()

Or the entire process can be run in a single line with

>>> w = WfcWrapper('jc8m32j5q_flc.fits', preprocess=True, extension=4, binsize=2,
...                execute=True)

"""  # noqa
import logging
import multiprocessing
import os
import warnings

import numpy as np
from astropy.io import fits
from astropy.nddata import bitmask, block_reduce
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename

from acstools.utils_findsat_mrt import (create_mask, filter_sources,
                                        merge_tables, radon, streak_endpoints,
                                        update_dq)

# test for matplotlib, turn off plotting if it does not exist
try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ImportError:
    plt = None
    warnings.warn('matplotlib not found, plotting is disabled')

__taskname__ = "findsat_mrt"
__author__ = "David V. Stark"
__version__ = "1.0"
__vdate__ = "10-Feb-2023"
__all__ = ['TrailFinder', 'WfcWrapper']

# Initialize the logger
logging.basicConfig()
LOG = logging.getLogger(f'{__taskname__}')
LOG.setLevel(logging.INFO)


class TrailFinder:
    '''Top-level class to handle trail identification and masking.

    Parameters
    ----------
    image : ndarray
        Input image.
    header : `astropy.io.fits.Header`, optional
        The primary header for the input data (usually the 0th extension). This
        is not used for anything during the analysis, but it is saved with the
        output mask and satellite trail catalog so information about the
        original observation can be easily retrieved. Default is `None`.
    image_header : `astropy.io.fits.Header`, optional
        The specific header for the FITS extension being used. This is
        added onto the catalog. Default is `None`.
    save_image_header_keys : list, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.save_image_header_keys`.
        Default is an empty list (nothing saved from ``image_header``).
    processes : int, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.processes`.
        The default is 2.
    min_length : int, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.min_length`.
        The default is 25 pixels.
    max_width : int, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.max_width`.
        The default is 75 pixels.
    buffer : int, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.buffer`.
        The default is 250 pixels on each side.
    threshold : float, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.threshold`.
        The default is 5.
    theta : ndarray, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.theta`.
        The default is `None`, which sets to ``numpy.arange(0, 180, 0.5)``.
    kernels : list, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.kernels`.
        The default is `None`, which reverts to using the 3-, 7-, and 15-pixel
        wide kernels included with this package.
    mask_include_status : list, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.mask_include_status`.
        The default is ``[2]``.
    plot : bool, optional
        Plot all intermediate steps. When set, plots of the input image, MRT
        with identified sources, and resulting image mask will be generated
        after running the constructor,
        :func:`acstools.findsat_mrt.TrailFinder.run_mrt`,
        :func:`acstools.findsat_mrt.TrailFinder.find_mrt_sources`,
        :func:`acstools.findsat_mrt.TrailFinder.filter_sources`,
        and :func:`acstools.findsat_mrt.TrailFinder.make_mask`. Users may also
        generate these plots manually by calling
        :func:`acstools.findsat_mrt.TrailFinder.plot_image`,
        :func:`acstools.findsat_mrt.TrailFinder.plot_mrt`,
        :func:`acstools.findsat_mrt.TrailFinder.plot_mask`, and
        :func:`acstools.findsat_mrt.TrailFinder.plot_segment`. The default is
        `False`.
    output_dir : str, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.output_dir`.
        The default is ``'.'`` (current directory).
    output_root : string, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.root`.
        The default is ``''`` (no prefix).
    check_persistence : bool, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.check_persistence`.
        The default is `True`.
    min_persistence : float, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.min_persistence`.
        Must be between 0 and 1. The default is 0.5.
    ignore_theta_range : list of tuples or `None`, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.ignore_theta_range`.
        Default is `None`.
    save_catalog : bool, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.save_catalog`.
        Default is `True`.
    save_diagnostic : bool, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.save_diagnostic`.
        Default is `True`.
    save_mrt : bool, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.save_mrt`.
        Default is `False`.
    save_mask : bool, optional
        See :attr:`~acstools.utils_findsat_mrt.TrailFinder.save_mask`.
        Default is `False`.

    '''  # noqa
    def __init__(
            self,
            image,
            header=None,
            image_header=None,
            save_image_header_keys=None,
            processes=2,
            min_length=25,
            max_width=75,
            buffer=250,
            threshold=5,
            theta=None,
            kernels=None,
            mask_include_status=[2],
            plot=False,
            output_dir='.',
            output_root='',
            check_persistence=True,
            min_persistence=0.5,
            ignore_theta_range=None,
            save_catalog=True,
            save_diagnostic=True,
            save_mrt=False,
            save_mask=False):

        # inputs
        self.image = image
        self.header = header
        self.image_header = image_header
        self.save_image_header_keys = save_image_header_keys
        self.threshold = threshold
        self.min_length = min_length
        self.max_width = max_width
        self.kernels = kernels
        self.processes = processes
        self.theta = theta
        self.plot = plot
        self.buffer = buffer
        self.check_persistence = check_persistence
        self.min_persistence = min_persistence
        self.mask_include_status = mask_include_status
        self.ignore_theta_range = ignore_theta_range

        # outputs
        self.mrt = None
        self.mrt_err = None
        self.length = None
        self.rho = None
        self.mask = None

        # some internal things
        self._madrt = None
        self._medrt = None
        self._image_mad = None
        self._image_stddev = None
        if plt is not None:
            self._interactive = mpl.is_interactive()
        else:
            self._interactive = False

        # info for what output to save and where
        self.output_dir = output_dir
        self.root = output_root
        self.save_catalog = save_catalog
        self.save_diagnostic = save_diagnostic
        self.save_mrt = save_mrt
        self.save_mask = save_mask

        # plot image upon initialization
        if self.plot and (plt is not None):
            self.plot_image()

    @property
    def save_image_header_keys(self):
        """List of header keys from ``self.image_header`` to save in the output
        trail catalog header.
        """
        return self._save_image_header_keys

    @save_image_header_keys.setter
    def save_image_header_keys(self, value):
        if value is None:
            value = []
        elif not isinstance(value, (list, tuple)):
            raise ValueError(f"save_image_header_keys must be list or tuple but got: {value}")  # noqa
        else:
            value = np.squeeze(value).tolist()
        self._save_image_header_keys = value

    @property
    def processes(self):
        """Number of processes to use when calculating MRT."""
        return self._processes

    @processes.setter
    def processes(self, value):
        max_cpu = multiprocessing.cpu_count()
        if value < 1:
            self._processes = 1
        elif value > max_cpu:
            self._processes = max_cpu
        else:
            self._processes = value

    @property
    def min_length(self):
        """Minimum streak length allowed.
        This is the minimum allowed length of a satellite trail.
        """
        return self._min_length

    @min_length.setter
    def min_length(self, value):
        if value < 1:
            raise ValueError(f"Invalid min_length: {value}")
        self._min_length = value

    @property
    def max_width(self):
        """Maximum streak width allowed.
        This is the maximum width of a trail to be considered robust.
        """
        return self._max_width

    @max_width.setter
    def max_width(self, value):
        if value < 1:
            raise ValueError(f"Invalid max_width: {value}")
        self._max_width = value

    @property
    def buffer(self):
        """Size of cutout region extending perpendicularly outward
        on each side from a streak/trail when analyzing its properties.
        """
        return self._buffer

    @buffer.setter
    def buffer(self, value):
        if value < 1:
            raise ValueError(f"Invalid buffer: {value}")
        self._buffer = value

    @property
    def threshold(self):
        """Minimum SNR when extracting sources from the MRT."""
        return self._threshold

    @threshold.setter
    def threshold(self, value):
        if value <= 0:
            raise ValueError(f"SNR threshold must be positive but got: {value}")  # noqa
        self._threshold = value

    @property
    def theta(self):
        """Angles at which to calculate the MRT."""
        return self._theta

    @theta.setter
    def theta(self, value):
        # NOTE: Not all assumptions used in find_mrt_sources method are
        # enforced here.
        if value is None:
            self._theta = np.arange(0, 180, 0.5)
        elif len(value) > 1:
            self._theta = value
        else:
            raise ValueError(f"Invalid theta: {value}")

    @property
    def kernels(self):
        """Paths to each kernel to be used for source finding in the MRT."""
        return self._kernels

    @kernels.setter
    def kernels(self, value):
        if value is None:
            self._kernels = [
                get_pkg_data_filename(os.path.join('data', f'rt_line_kernel_width{k}.fits'))  # noqa
                for k in (15, 7, 3)]
        else:
            user_files = [v for v in value if os.path.isfile(v)]
            if len(user_files) < 1:
                raise ValueError(f"No usable kernel provided: {value}")
            self._kernels = user_files

    @property
    def mask_include_status(self):
        """List indicating trails with which status should be considered
        when making the mask. Statuses are generated by
        :func:`acstools.utils_findsat_mrt.filter_sources`:

        * 1 = Failed SNR or width requirements.
        * 2 = Passed SNR and width requirements but failed persistence test.
        * 3 = Passed SNR, width, and persistence requirements.
        """
        return self._mask_include_status

    @mask_include_status.setter
    def mask_include_status(self, value):
        if (not isinstance(value, list)) or (set(value) > {1, 2, 3}):
            raise ValueError(f"Status list elements must be 1, 2, or 3 but got: {value}")  # noqa
        self._mask_include_status = value

    @property
    def output_dir(self):
        """Path in which to save output."""
        return self._output_dir

    @output_dir.setter
    def output_dir(self, value):
        if not os.path.isdir(value):
            raise ValueError(f"output_dir must be a directory but got: {value}")  # noqa
        self._output_dir = value

    @property
    def root(self):
        """A prefix for all output files."""
        return self._root

    @root.setter
    def root(self, value):
        if not isinstance(value, str):
            raise ValueError(f"root must be a string but got: {value}")
        self._root = value

    @property
    def check_persistence(self):
        """Calculate the persistence of all identified streaks."""
        return self._check_persistence

    @check_persistence.setter
    def check_persistence(self, value):
        if not isinstance(value, bool):
            raise ValueError(f"check_persistence must be bool but got: {value}")  # noqa
        self._check_persistence = value

    @property
    def min_persistence(self):
        """Minimum persistence of a "true" satellite trail to be considered
        robust. Note that this does not reject satellite trails from the output
        catalog, but highlights them in a different color in the output plot.
        """
        return self._min_persistence

    @min_persistence.setter
    def min_persistence(self, value):
        if value < 0 or value > 1:
            raise ValueError(f"min_persistence must be between 0 and 1 but got: {value}")  # noqa
        self._min_persistence = value

    @property
    def ignore_theta_range(self):
        """List of ranges in `theta` to ignore when identifying satellite
        trails. This parameter is most useful for avoiding false positives
        due to diffraction spikes that always create streaks around the
        same angle for a given telescope/instrument. Format should be a
        list of tuples, e.g., ``[(theta0_a, theta1_a), (theta0_b, theta1_b)]``.
        """
        return self._ignore_theta_range

    @ignore_theta_range.setter
    def ignore_theta_range(self, value):
        if ((value is None) or
                (isinstance(value, list) and all(len(x) == 2 for x in value))):
            self._ignore_theta_range = value
        else:
            raise ValueError(f"Invalid ignore_theta_range: {value}")

    @property
    def save_catalog(self):
        """Save the catalog of identified trails to a FITS table."""
        return self._save_catalog

    @save_catalog.setter
    def save_catalog(self, value):
        if not isinstance(value, bool):
            raise ValueError(f"save_catalog must be bool but got: {value}")
        self._save_catalog = value

    @property
    def save_diagnostic(self):
        """Save a diagnotic plot showing the input image and identified trails
        to a PNG file."""
        return self._save_diagnostic

    @save_diagnostic.setter
    def save_diagnostic(self, value):
        if not isinstance(value, bool):
            raise ValueError(f"save_diagnostic must be bool but got: {value}")
        self._save_diagnostic = value

    @property
    def save_mrt(self):
        """Save the MRT in a FITS file."""
        return self._save_mrt

    @save_mrt.setter
    def save_mrt(self, value):
        if not isinstance(value, bool):
            raise ValueError(f"save_mrt must be bool but got: {value}")
        self._save_mrt = value

    @property
    def save_mask(self):
        """Save the trail mask in a FITS file."""
        return self._save_mask

    @save_mask.setter
    def save_mask(self, value):
        if not isinstance(value, bool):
            raise ValueError(f"save_mask must be bool but got: {value}")
        self._save_mask = value

    def run_mrt(self):
        '''
        Run the median radon transform on the input image.
        This uses `theta` and `processes` for calculations, so update them
        first, if needed.

        '''
        # run MRT
        rt, length = radon(self.image, circle=False, median=True,
                           fill_value=np.nan, processes=self.processes,
                           return_length=True, theta=self.theta)

        # trim mrt where length too short
        rt[length < self.min_length] = np.nan

        # retain various outputs
        self.mrt = rt
        self.length = length

        # setting a warning filter for the nanmedian calculations. These
        # warnings are inconsequential and already accounted for in the code
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore',
                                    message='All-NaN slice encountered')
            # calculate some useful properties
            # median
            self._medrt = np.nanmedian(rt)
            # median abs deviation
            self._madrt = np.nanmedian(np.abs(rt[np.abs(rt) > 0]) -
                                       self._medrt)

            # calculate the approximate uncertainty of the MRT at each point
            self._image_mad = np.nanmedian(np.abs(self.image))

        # using MAD to avoid influence from outliers
        self._image_stddev = self._image_mad / 0.67449
        # error on median ~ 1.25x error on mean. There are regions with length
        # equals zero which keeps raising warnings. Suppressing that warning
        # here
        with np.errstate(divide='ignore', invalid='ignore'):
            self.mrt_err = 1.25 * self._image_stddev / np.sqrt(self.length)

        # create the rho array
        rho0 = rt.shape[0] / 2 - 0.5
        self.rho = np.arange(rt.shape[0]) - rho0

        # plot if set
        if self.plot and (plt is not None):
            self.plot_mrt()

    def plot_image(self, ax=None, scale=(-1, 5), overlay_mask=False):
        '''
        Plot the input image.

        Parameters
        ----------
        ax : `matplotlib.axes.Axes`, optional
            A matplotlib subplot where the image should be shown. The default
            is `None` (one will be created for you).
        scale : tuple of floats, optional
            A two element array with the minimum and maximum image values used
            to set the color scale, in units of the image median absolute
            deviation (MAD). The default is ``(-1, 5)``.
        overlay_mask : bool, optional
            Overlay the trail mask, if already calculated. Default is `False`.

        '''
        if plt is None:
            return

        if ax is None:
            fig, ax = plt.subplots()

        # recalculate mad and stdev here in case it hasn't been done yet
        # setting a warning filter for the nanmedian calculations. These
        # warnings are inconsequential and already accounted for in the code
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore',
                                    message='All-NaN slice encountered')
            self._image_mad = np.nanmedian(np.abs(self.image))

        self._image_stddev = self._image_mad / 0.67449  # using MAD to avoid
        # influence from outliers

        ax.imshow(self.image, cmap='viridis', origin='lower', aspect='auto',
                  vmin=scale[0]*self._image_stddev,
                  vmax=scale[1]*self._image_stddev)
        ax.set_xlabel('X [pix]')
        ax.set_ylabel('Y [pix]')
        ax.set_title('Input Image')

        if overlay_mask is True:
            if self.mask is None:
                LOG.error('No mask to overlay')
            else:
                ax.imshow(self.mask, alpha=0.5, cmap='Reds', origin='lower',
                          aspect='auto')

        # write to file if interactive is off and plottin is on
        if self.plot and ~self._interactive:
            file_name = os.path.join(self.output_dir, self.root + '_image')
            if overlay_mask:
                file_name = file_name + '_mask'
            plt.savefig(file_name + '.png')

    def plot_mrt(self, ax=None, scale=(0, 10), show_sources=False):
        '''
        Plot the MRT.

        Parameters
        ----------
        ax : `matplotlib.axes.Axes`, optional
            A matplotlib subplot where MRT should be shown. The default
            is `None` (one will be created for you).
        scale : tuple of floats, optional
            A two element array with the minimum and maximum image values used
            to set the color scale, in units of the MRT median absolute
            deviation (MAD). The default is ``(-1, 5)``.
        show_sources : bool
            Mark the positions of the detected sources. Default is `False`.

        Returns
        -------
        ax : `matplotlib.axes.Axes`
            Matplotlib subplot where the MRT is plotted.

        '''
        # exit if no MRT
        if self.mrt is None:
            LOG.error('No MRT to plot')
            return

        # Letting user know if there are no sources to overplot
        if show_sources and (self.source_list is None):
            show_sources = False
            LOG.info('No sources to show')

        if ax is None:
            fig, ax = plt.subplots()

        ax.imshow(self.mrt, aspect='auto', origin='lower',
                  vmin=scale[0]*self._madrt, vmax=scale[1]*self._madrt)
        ax.set_title('MRT')
        ax.set_xlabel('angle(theta) pixel')
        ax.set_ylabel('offset(rho) pixel')

        # overplot sources if applicable
        if show_sources is True:
            x = self.source_list['xcentroid']
            y = self.source_list['ycentroid']
            status = self.source_list['status']

            for s, color in zip([0, 1, 2], ['red', 'orange', 'cyan']):
                sel = (status == s)
                if np.sum(sel) > 0:
                    ax.scatter(x[sel], y[sel], edgecolor=color,
                               facecolor='none', s=100, lw=2,
                               label=f'status={s}')
            ax.legend(loc='upper center')

        # send plot to file if interactive is off
        if self.plot and ~self._interactive:
            file_name = os.path.join(self.output_dir, self.root + '_mrt')
            if show_sources:
                file_name = file_name + '_sources'
            file_name = file_name + '.png'
            plt.savefig(file_name)
            LOG.info('Saving MRT to '+file_name)

        return ax

    def plot_mrt_snr(self, ax=None, scale=(1, 25)):
        '''
        Plots a map of the MRT signal-to-noise ratio (SNR).

        Parameters
        ----------
        ax : `matplotlib.axes.Axes`, optional
            A matplotlib subplot where the SNR should be shown. The default
            is `None` (one will be created for you).
        scale : tuple of floats, optional
            A two element array with the minimum and maximum image values used
            to set the color scale. The default is ``(1, 25)``.

        Returns
        -------
        snr_map : array-like
            A map of the SNR.

        '''
        if self.mrt is None:
            LOG.error('No MRT to plot')
            return

        if ax is None:
            fig, ax = plt.subplots()

        snr = self.mrt / self.mrt_err
        ax.imshow(snr, aspect='auto', origin='lower',
                  vmin=scale[0], vmax=scale[1])
        ax.set_title('MRT SNR')
        ax.set_xlabel('angle(theta) pixel')
        ax.set_ylabel('offset(rho) pixel')

        if self.plot and ~self._interactive:
            file_name = os.path.join(self.output_dir, self.root +
                                     '_mrt_snr.png')
            plt.savefig(file_name)

        return snr

    def find_mrt_sources(self):
        '''
        Finds sources in the MRT consistent with satellite trails/streaks.
        This uses `kernels` and `threshold` for calculations, so update them
        first, if needed.

        Returns
        -------
        source_list : `~astropy.table.QTable` or `None`
            Catalog containing information about detected trails, if
            applicable. This is the same info as ``self.source_list``.

        '''
        # test for photutils
        try:
            from photutils.detection import StarFinder
        except ImportError:
            LOG.error('photutils not installed. Source detection will not work.')  # noqa
            return

        LOG.info('Detection threshold: {}'.format(self.threshold))

        snrmap = self.mrt / self.mrt_err
        snrmap_mask = ~np.isfinite(snrmap)

        # cycle through kernels
        tbls = []   # we'll store detected sources here
        for k in self.kernels:
            with fits.open(k) as h:
                kernel = h[0].data
                LOG.info('Using kernel {}'.format(k))

                # detect sources
                s = StarFinder(self.threshold, kernel, min_separation=20,
                               exclude_border=False, brightest=None,
                               peakmax=None)

            # can fail for cases where nothing found. Allow code to return
            # nothing and move on

            tbl = s.find_stars(snrmap, mask=snrmap_mask)
            if tbl is None:
                nsources = 0
            else:
                nsources = len(tbl)

            LOG.info('{{no}} sources found using kernel: {}'.format(nsources))

            if nsources > 0:
                tbl = tbl[np.isfinite(tbl['xcentroid'])]
                LOG.info('{} sources found using kernel'.format(len(tbl)))
                if (len(tbls) > 0):
                    if len(tbls[-1]['id']) > 0:
                        tbl['id'] += np.max(tbls[-1]['id'])
                        # adding max ID number from last iteration to avoid
                        # duplicate ids
                tbls.append(tbl)

        # combine tables from each kernel and remove any duplicates
        if len(tbls) > 0:
            LOG.info('Removing duplicate sources')
            sources = merge_tables(tbls)
            self.source_list = sources
        else:
            self.source_list = None

        # add the theta and rho values to the table
        if self.source_list is not None:
            dtheta = self.theta[1] - self.theta[0]
            self.source_list['theta'] = self.theta[0] + \
                dtheta * self.source_list['xcentroid']
            self.source_list['rho'] = self.rho[0] + \
                self.source_list['ycentroid']

            # Add the status array and endpoints array. Status will be zero
            # because no additional checks have been done yet.
            self.source_list['endpoints'] = [
                streak_endpoints(t['rho'], -t['theta'], self.image.shape)
                for t in self.source_list]
            self.source_list['status'] = np.zeros(len(self.source_list),
                                                  dtype=int)

            # run the routine to remove angles if any bad ranges are specified
            if self.ignore_theta_range is not None:
                self._remove_angles()

            # print the total number of sources found
            LOG.info('{} final sources found'.format(len(self.source_list)))
            # plot sources if set
            if self.plot and (plt is not None):
                self.plot_mrt(show_sources=True)
        else:
            LOG.warning('No sources found')

        return self.source_list

    def filter_sources(self, trim_catalog=False, plot_streak=False):
        '''
        Filters catalog of trails based on SNR, width, and persistence.
        This uses `threshold`, `max_width`, `min_length`, `buffer`,
        `check_persistence`, and `min_persistence` for calculations,
        so update them first, if needed.

        Parameters
        ----------
        trim_catalog : bool, optional
            Flag to remove all filtered trails from the source catalog. The
            default is `False`.
        plot_streak : bool, optional
            Set to plot diagnostics for each trail. Only works in interactive
            mode. Warning: this can generate a lot of plots depending on how
            many trails are found. Default is `False`.

        Returns
        -------
        source_list : `~astropy.table.QTable` or `None`
            Catalog of identified satellite trails with additional measured
            parameters appended. This is the same info as ``self.source_list``.

        '''
        if self.source_list is None:  # Nothing to do.
            return

        # check inputs
        if self.plot and ~self._interactive:
            plot_streak = False

        LOG.info('Filtering sources...\n'
                 'Min SNR : {}\n'
                 'Max Width: {}\n'
                 'Min Length: {}\n'
                 'Check persistence: {}'.format(
                     self.threshold, self.max_width, self.min_length,
                     self.check_persistence))

        if self.check_persistence is True:
            LOG.info('Min persistence: {}'.format(self.min_persistence))

        # run filtering routine
        properties = filter_sources(self.image,
                                    self.source_list['endpoints'],
                                    max_width=self.max_width,
                                    buffer=self.buffer,
                                    plot_streak=plot_streak,
                                    min_length=self.min_length,
                                    minsnr=self.threshold,
                                    check_persistence=self.check_persistence,
                                    min_persistence=self.min_persistence)

        # update the status
        self.source_list.update(properties)

        # optionally trim the catalog of all rejected sources
        if trim_catalog:
            sel = ((self.source_list['width'] < self.max_width) &
                   (self.source_list['snr'] > self.threshold))
            self.source_list = self.source_list[sel]

        # plot if triggered
        if self.plot and (plt is not None):
            self.plot_mrt(show_sources=True)

        return self.source_list

    def make_mask(self):
        '''
        Makes a satellite trail mask (bool) and a segmentation map.
        This uses `mask_include_status`, so update it first, if needed.

        The segmentation map is an image where pixels belonging to a given
        trail have values equal to the trail ID number.

        This updates ``self.segment`` and ``self.mask``.

        '''
        # crate the mask/segmentation
        if self.source_list is not None:
            include = [s['status'] in self.mask_include_status for s in
                       self.source_list]
            trail_id = self.source_list['id'][include]
            endpoints = self.source_list['endpoints'][include]
            widths = self.source_list['width'][include]
            segment, mask = create_mask(self.image, trail_id, endpoints,
                                        widths)
        else:
            mask = np.zeros(self.image.shape, dtype=bool)
            segment = np.zeros(self.image.shape, dtype=int)

        self.segment = segment
        self.mask = mask

        # plot if triggered
        if self.plot and (plt is not None):
            self.plot_mask()
            self.plot_segment()

    def plot_mask(self):
        '''
        Generates a plot of the trail mask (``self.mask``).

        Returns
        -------
        ax : `matplotlib.axes.Axes`
            The matplotlib subplot containing the mask image.

        '''
        # exit if there's not mask
        if self.mask is None:
            LOG.error('No mask to show')
            return

        fig, ax = plt.subplots()
        ax.imshow(self.mask, origin='lower', aspect='auto')
        ax.set_title('Mask')

        # send mask to file if interactive off and plotting enabled
        if self.plot and ~self._interactive:
            file_name = os.path.join(self.output_dir, self.root + '_mask.png')
            plt.savefig(file_name)
            LOG.info('Saving mask to '+file_name)

        return ax

    def plot_segment(self):
        '''
        Generates a segmentation image of the identified trails
        (``self.segment``).

        Returns
        -------
        ax : `matplotlib.axes.Axes`
            A matplotlib subplot containing the segmentation map.

        '''
        # exit if no segmentation image
        if self.segment is None:
            LOG.error('No segment map to show')
            return

        # get unique values in segment
        unique_vals = np.unique(self.segment)
        data = np.zeros_like(self.segment)
        counter = 1
        for uv in unique_vals[1:]:
            data[self.segment == uv] = counter
            counter += 1

        data_min = np.min(data).astype(int)
        data_max = np.max(data).astype(int)
        fig, ax = plt.subplots()

        # update the colormap to match the segmentation IDs
        cmap = plt.get_cmap('tab20', data_max - data_min + 1)
        mat = ax.imshow(data, cmap=cmap, vmin=data_min - 0.5,
                        vmax=data_max + 0.5, origin='lower', aspect='auto')

        # tell the colorbar to tick at integers
        ticks = np.arange(len(unique_vals) + 1)
        cax = plt.colorbar(mat, ticks=ticks)
        cax.ax.set_yticklabels(np.concatenate([unique_vals,
                                               [unique_vals[-1] + 1]]))
        cax.ax.set_ylabel('trail ID')
        ax.set_title('Segmentation Mask')

        # send to file if not interactive
        if self.plot and ~self._interactive:
            file_name = os.path.join(self.output_dir, self.root +
                                     '_segment.png')
            plt.savefig(file_name)
            LOG.info('Saving segmentation map to '+file_name)

        return ax

    def save_output(self, close_plot=True):
        '''
        Save output. Any existing file will be overwritten.
        This uses `root`, `output_dir`, `save_mrt`, `save_catalog`, and
        `save_diagnostic`, so update them first, if needed. Output includes
        optionally:

        1. MRT
        2. Mask/segementation image
        3. Catalog
        4. Trail catalog

        Parameters
        ----------
        close_plot : bool
            Close the plot instance after `save_diagnostic` is done.

        '''
        # save the MRT image
        if self.save_mrt:
            if self.mrt is not None:
                mrt_filename = os.path.join(self.output_dir,
                                            f'{self.root}_mrt.fits')
                fits.writeto(mrt_filename, self.mrt, overwrite=True)
                LOG.info('Wrote MRT to {}'.format(mrt_filename))
            else:
                LOG.error('No MRT to save')

        # save the bool mask image
        if self.save_mask:
            if self.mask is not None and self.segment is not None:
                hdu0 = fits.PrimaryHDU()
                if self.header is not None:
                    hdu0.header = self.header  # copying original image header
                hdu1 = fits.ImageHDU(self.mask.astype(int))
                hdu2 = fits.ImageHDU(self.segment.astype(int))
                hdul = fits.HDUList([hdu0, hdu1, hdu2])
                mask_filename = os.path.join(self.output_dir,
                                             f'{self.root}_mask.fits')
                hdul.writeto(mask_filename, overwrite=True)
                LOG.info('Wrote mask to {}'.format(mask_filename))
            else:
                LOG.error('No mask to save')

        # save the diagnostic plot
        if self.save_diagnostic and (plt is not None):

            fig, [[ax1, ax2], [ax3, ax4]] = plt.subplots(2, 2,
                                                         figsize=(20, 10))
            self.plot_image(ax=ax1)
            self.plot_mrt(ax=ax2)
            self.plot_image(ax=ax3)
            ax3.imshow(self.mask, alpha=0.5, origin='lower', aspect='auto',
                       cmap='Reds')
            ax3_xlim = ax3.get_xlim()
            ax3_ylim = ax3.get_ylim()

            self.plot_mrt(ax=ax4)
            if self.source_list is not None:
                for s in self.source_list:
                    color = 'red'
                    x1, y1 = s['endpoints'][0]
                    x2, y2 = s['endpoints'][1]
                    if (s['status'] == 2):
                        color = 'turquoise'
                    elif (s['status'] == 1):
                        color = 'orange'
                    ax3.plot([x1, x2], [y1, y2], color=color, lw=5, alpha=0.5)
                    ax4.scatter(s['xcentroid'], s['ycentroid'],
                                edgecolor=color, facecolor='none', s=100, lw=2)
            # sometimes overplotting the "good" trails can cause axes to change
            ax3.set_xlim(ax3_xlim)
            ax3.set_ylim(ax3_ylim)
            plt.tight_layout()
            diagnostic_filename = os.path.join(self.output_dir,
                                               f'{self.root}_diagnostic.png')
            plt.savefig(diagnostic_filename)
            LOG.info('Wrote diagnostic plot to {}'.format(diagnostic_filename))
            if close_plot:
                plt.close()

        # save the catalog of trails
        if self.save_catalog:
            cat_filename = os.path.join(self.output_dir,
                                        f'{self.root}_catalog.fits')
            if self.source_list is not None:
                # there's some meta data called "version" that cannot be added
                # to the fits header (and it is not useful). It always throws
                # an inconsequential warning so we suppress it here.
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore',
                                            message='Attribute `version` of.*')
                    self.source_list.write(cat_filename, overwrite=True)
            else:
                # create an empty catalog and write that. It helps to have this
                # for future analysis purposes even if it's empty
                dummy_table = Table(names=('id', 'xcentroid', 'ycentroid',
                                           'fwhm', 'roundness', 'pa',
                                           'max_value', 'flux', 'mag',
                                           'theta', 'rho', 'endpoints',
                                           'width', 'snr', 'status',
                                           'persistence'),
                                    dtype=('int64', 'float64', 'float64',
                                           'float64', 'float64', 'float64',
                                           'float64', 'float64', 'float64',
                                           'float64', 'float64', 'float64',
                                           'float64', 'float64', 'float64',
                                           'float64'))
                dummy_table.write(cat_filename, overwrite=True)
            LOG.info(f'wrote catalog {cat_filename}')

            # Append the original data header to this catalog too.
            has_pri_hdr = isinstance(self.header, fits.Header)
            has_dat_hdr = (isinstance(self.image_header, fits.Header) and
                           (len(self.save_image_header_keys) > 0))
            if has_pri_hdr or has_dat_hdr:
                with fits.open(cat_filename, mode='update') as h:
                    if has_pri_hdr:
                        h[0].header.update(self.header)

                    if has_dat_hdr:
                        # add individal header keywords now
                        for k in self.save_image_header_keys:
                            if k in self.image_header:
                                h[1].header[k] = self.image_header[k]

    def _remove_angles(self):
        '''
        Remove a range (or set of ranges) of angles from the trail catalog
        by updating ``self.source_list`` in-place.
        This uses `ignore_theta_range`, so update it first, if needed.

        This routine is primarily for removing trails at angles known to
        be overwhelmingly dominated by features that are not of interest, e.g.,
        for removing diffraction spikes.

        '''
        if self.source_list is None:  # Nothing to do
            return

        if self.ignore_theta_range is None:
            LOG.error('No angles set to ignore')
            return

        # add some checks to be sure ignore_ranges is the right type
        remove = np.zeros(len(self.source_list), dtype=bool)
        for r in self.ignore_theta_range:
            r = np.sort(r)
            LOG.info('ignoring angles between {} and {}'.format(r[0], r[1]))
            remove[(self.source_list['theta'] >= r[0]) &
                   (self.source_list['theta'] <= r[1])] = True

        self.source_list = self.source_list[~remove]

    def run_all(self, trim_catalog=False, plot_streak=False, close_plot=True):
        '''
        Run the entire pipeline to identify, filter, and mask trails.
        This calls the following methods in the given order:

        1. :meth:`run_mrt`
        2. :meth:`find_mrt_sources`
        3. :meth:`filter_sources`
        4. :meth:`make_mask`
        5. :meth:`save_output`

        See the documentation for methods above on how to use the
        keyword options.

        '''
        self.run_mrt()
        self.find_mrt_sources()
        self.filter_sources(trim_catalog=trim_catalog, plot_streak=plot_streak)
        self.make_mask()
        self.save_output(close_plot=close_plot)


class WfcWrapper(TrailFinder):
    '''
    Wrapper for `TrailFinder` designed specifically for ACS/WFC data.

    This class enables quick reading and preprocessing of standard
    full-frame ACS/WFC images.

    .. note::

        This class is not designed for use with subarray images.

    Parameters
    ----------
    image_file : str
        ACS/WFC data file to read. Should be a FITS file.
    extension : int, optional
        Extension of input file to read. This keyword is required if the input
        image is an FLC/FLT file. The default is `None`.
    binsize : int or `None`, optional
        See :attr:`~acstools.utils_findsat_mrt.WfcWrapper.binsize`.
        The default is `None` (no binning).
    ignore_flags : list of int
        See :attr:`~acstools.utils_findsat_mrt.WfcWrapper.ignore_flags`.
        Default is ``[4096, 8192, 16384]``.
    preprocess : bool, optional
        Flag to run all the preprocessing steps (bad pixel flagging,
        background subtraction, rebinning). The default is `True`.
    execute : bool, optional
        Flag to run the entire `TrailFinder` pipeline. The default is `False`.
    **kwargs : dict, optional
        Additional keyword arguments for `TrailFinder`.

    Raises
    ------
    ValueError
        Image is subarray, or unrecognized image extension or type.

    '''
    def __init__(self,
                 image_file,
                 extension=None,
                 binsize=None,
                 ignore_flags=[4096, 8192, 16384],
                 preprocess=True,
                 execute=False,
                 **kwargs):
        self.image_file = image_file
        self.extension = extension
        self.binsize = binsize
        self.ignore_flags = ignore_flags
        self.preprocess = preprocess
        self.execute = execute

        # open image
        with fits.open(self.image_file) as h:

            # check that image ID not subarray
            if h[0].header['SUBARRAY'] is True:
                raise ValueError('This program does not yet work on subarrays')

            # get suffix to determine how to process image
            suffix = (self.image_file.split('.')[-2]).split('_')[-1].lower()
            self.image_type = suffix
            LOG.info('image type is {}'.format(self.image_type))

            if suffix in ('flc', 'flt'):
                if extension is None:
                    raise ValueError('FLC/FLT files must have extension specified')  # noqa
                elif extension not in (1, 4):
                    raise ValueError('Valid extensions are 1 or 4 for FLC/FLT'
                                     f'images but got: {extension}')

                image = h[extension].data  # main image
                self.image_mask = h[extension + 2].data  # DQ array

            elif suffix in ('drc', 'drz'):
                extension = 1
                image = h[extension].data  # main image
                self.image_mask = h[extension + 1].data  # weight array

            else:
                raise ValueError(f'Image type not recognized: {suffix}')

        super().__init__(image, **kwargs)

        # go ahead and run the proprocessing steps if set to True
        if preprocess:
            self.run_preprocess()

        # run the whole pipeline if set to True
        if execute:
            LOG.info('Running the trailfinding pipeline')
            self.run_all()

    @property
    def binsize(self):
        """Amount the input data should be binned by."""
        return self._binsize

    @binsize.setter
    def binsize(self, value):
        if value is not None and not (isinstance(value, int) and value >= 1):
            raise ValueError(f"binsize must be None or int >= 1 but got: {value}")  # noqa
        self._binsize = value

    @property
    def ignore_flags(self):
        """DQ flags that are ignored when flagging bad pixels.
        Only relevant for FLT/FLC files.
        """
        return self._ignore_flags

    @ignore_flags.setter
    def ignore_flags(self, value):
        if not isinstance(value, list) or not all(isinstance(v, int) for v in
                                                  value):
            raise ValueError(f"ignore_flags must be list of int but got:"
                             f"{value}")
        self._ignore_flags = value

    def mask_bad_pixels(self):
        '''
        Mask bad pixels.
        This uses `ignore_flags` (for FLC/FLT), so update it first, if needed.

        Bad pixels are replaced with NaN by modifying ``self.image`` in-place.
        This uses the bitmask arrays for FLC/FLT images,
        and weight arrays for DRC/DRZ images.

        '''
        LOG.info('masking bad pixels')

        if self.image_type in ('flc', 'flt'):
            # for FLC/FLT, use DQ array
            mask = bitmask.bitfield_to_boolean_mask(self.image_mask,
                                                    ignore_flags=self.ignore_flags)  # noqa
            self.image[mask] = np.nan

        elif self.image_type in ('drz', 'drc'):
            # for DRZ/DRC, mask everything with weight=0
            self.image[self.image_mask == 0] = np.nan

    def subtract_background(self):
        '''
        Subtract a median background from the image, ignoring NaNs,
        where ``self.image`` is modified in-place.
        Bad pixels should first be handled via :meth:`mask_bad_pixels`
        before running this.
        '''
        LOG.info('Subtracting median background')

        # setting a warning filter for the nanmedian calculations. These
        # warnings are inconsequential and already accounted for in the code
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore',
                                    message='All-NaN slice encountered')
            self.image = self.image - np.nanmedian(self.image)

    def rebin(self):
        '''
        Rebin the image array by modifying ``self.image`` in-place.
        The binning in the x and y direction are identical.
        This uses `binsize`, so update it first, if needed.

        '''
        if self.binsize is None:
            LOG.warn('No bin size defined. Will not perform binning')
            return

        LOG.info('Rebinning the data by {}'.format(self.binsize))

        # trigger a warning if the image dimensions are not even
        if (((self.image.shape[0] % self.binsize != 0)) |
           ((self.image.shape[1] % self.binsize != 0))):
            LOG.warning('Image dimensions do not evenly divide by bin size.'
                        '\nExtra rows/columns will be trimmed first.')

        # setting a warning filter for the nansum calculations. These
        # warnings are inconsequential and already accounted for in the code
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore',
                                    message='All-NaN slice encountered')
            self.image = block_reduce(self.image, self.binsize, func=np.nansum)

    def run_preprocess(self):
        '''
        Runs all the image preprocessing steps in the following order:

        1. :meth:`mask_bad_pixels`
        2. :meth:`subtract_background`
        3. :meth:`rebin`

        See the documentation for methods above for more details.

        '''
        self.mask_bad_pixels()
        self.subtract_background()
        self.rebin()

    def update_dq(self, dqval=16384, verbose=True):
        '''
        Update DQ array with the satellite trail mask (``self.mask``) using
        :func:`acstools.utils_findsat_mrt.update_dq`.
        The file in ``self.image_file`` is updated for this operation,
        particularly the DQ extension associated with ``self.extension``.

        .. note::

            This routine only works on FLC/FLT images.

        Parameters
        ----------
        dqval : int, optional
            DQ value to use for the trail. Default value of 16384 is
            tailored for ACS/WFC.
        verbose : bool, optional
            Print extra information to the terminal.

        '''
        if self.image_type not in ('flc', 'flt'):
            raise ValueError(f'DQ array can only be updated for FLC/FLT images, not {self.image_type}')  # noqa

        update_dq(self.image_file, self.extension + 2, self.mask, dqval=dqval,
                  verbose=verbose, expand_mask=True)
